"""
Phase labeling tools for clustering and classification of materials data.

This module provides classes for automatically identifying and labeling phases in
materials science data. It includes various clustering algorithms and methods for
determining optimal number of phases using silhouette analysis.

Key features:
- Multiple clustering algorithms (Spectral, GMM, Affinity Propagation)
- Automatic determination of number of phases via Silhouette analysis
- Support for precomputed similarity/distance matrices
- Integration with scikit-learn clustering algorithms
"""

import copy
import warnings
from collections import Counter
from collections import defaultdict

import numpy as np
import xarray as xr
import sklearn.cluster
import sklearn.mixture
from sklearn.metrics import silhouette_samples

from AFL.double_agent.PipelineOp import PipelineOp


class Labeler(PipelineOp):
    """Base class for phase labeling operations.
    
    This abstract base class provides common functionality for labeling phases
    in materials data. It supports various clustering approaches and includes
    methods for label manipulation and silhouette analysis.

    Parameters
    ----------
    input_variable : str
        The name of the variable containing the data to be labeled

    output_variable : str
        The name of the variable where labels will be stored

    dim : str, default='sample'
        The dimension name for samples in the dataset

    use_silhouette : bool, default=False
        Whether to use silhouette analysis to determine optimal number of phases

    params : dict, optional
        Additional parameters for the labeling algorithm

    name : str, default='PhaseLabeler'
        The name to use when added to a Pipeline
    """

    def __init__(self, input_variable, output_variable, dim='sample', use_silhouette=False, params=None, name='PhaseLabeler'):
        super().__init__(name=name, input_variable=input_variable, output_variable=output_variable)

        self.labels = None
        self.dim = dim

        if params is None:
            self.params = {}
        else:
            self.params = params

        self._banned_from_attrs.extend(['labels','clf','silh_dict','params'])
        self.use_silhouette = use_silhouette

    def __getitem__(self, index):
        """Get labels at specified index."""
        return self.labels[index]

    def __array__(self, dtype=None):
        """Convert labels to numpy array."""
        return np.array(self.labels).astype(dtype)

    def remap_labels_by_count(self):
        """Remap phase labels to be ordered by frequency.
        
        Reorders labels so that the most common phase is labeled 0,
        second most common is 1, etc.
        """
        label_map = {}
        for new_label, (old_label, _) in enumerate(
                sorted(Counter(self.labels).items(), key=lambda x: x[1], reverse=True)):
            label_map[old_label] = new_label
        self.labels = list(map(label_map.get, self.labels))

    def silhouette(self, W):
        """Perform silhouette analysis to determine optimal number of phases.
        
        Uses silhouette scores to automatically determine the best number of
        phases by trying different numbers of clusters and analyzing the
        clustering quality.

        Parameters
        ----------
        W : array-like
            The similarity/distance matrix to use for clustering

        Notes
        -----
        This method modifies the instance's n_phases and labels attributes
        based on the silhouette analysis results.
        """
        X = W.copy()
        silh_dict = defaultdict(list)
        max_n = min(X.shape[0], 15)
        for n_phases in range(2, max_n):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._construct_labeler(n_phases=n_phases)
                self._label(X)
            self.remap_labels_by_count()

            if len(np.unique(self.labels)) == 1:
                silh_scores = np.zeros(len(X))
            else:
                silh_scores = silhouette_samples(
                    1.0 - X,
                    self,
                    metric='precomputed'
                )
            silh_dict['all_scores'].append(silh_scores)
            silh_dict['avg_scores'].append(silh_scores.mean())
            silh_dict['n_phases'].append(n_phases)
            silh_dict['labels'].append(self.labels.copy())

        silh_avg = np.array(silh_dict['avg_scores'])
        found = False
        for cutoff in np.arange(0.85, 0.05, -0.05):
            idx = np.where(silh_avg > cutoff)[0]
            if idx.shape[0] > 0:
                idx = idx[-1]
                found = True
                break

        if not found:
            self.n_phases = 1
            self.labels = np.zeros(X.shape[0])
        else:
            self.n_phases = silh_dict['n_phases'][idx]
            self.labels = silh_dict['labels'][idx]
        self.silh_dict = silh_dict


class SpectralClustering(Labeler):
    """Spectral clustering for phase identification.
    
    Uses spectral clustering to identify phases based on a similarity matrix.
    Particularly effective for non-spherical clusters and when working with
    similarity rather than distance metrics.

    Parameters
    ----------
    input_variable : str
        The name of the variable containing the similarity matrix

    output_variable : str
        The name of the variable where labels will be stored

    dim : str
        The dimension name for samples in the dataset

    params : dict, optional
        Additional parameters for spectral clustering

    name : str, default='SpectralClustering'
        The name to use when added to a Pipeline

    use_silhouette : bool, default=False
        Whether to use silhouette analysis to determine optimal number of phases
    """

    def __init__(self, input_variable, output_variable, dim, params=None, name='SpectralClustering', use_silhouette=False):
        super().__init__(name=name, input_variable=input_variable, output_variable=output_variable, dim=dim,
                         params=params, use_silhouette=use_silhouette)

    def _construct_labeler(self, **params):
        """Construct the spectral clustering model.
        
        Parameters
        ----------
        **params : dict
            Additional parameters to update the clustering configuration
        """
        self.params.update(params)

        self.clf = sklearn.cluster.SpectralClustering(
            self.params['n_phases'],
            affinity='precomputed',
            assign_labels="discretize",
            random_state=0,
            n_init=1000
        )

    def calculate(self, dataset):
        """Apply spectral clustering to the dataset.
        
        Parameters
        ----------
        dataset : xr.Dataset
            The input dataset containing the similarity matrix

        Returns
        -------
        Self
            The SpectralClustering instance with computed labels
        """
        data1 = self._get_variable(dataset)
        self._construct_labeler()
        if self.use_silhouette:
            self.silhouette(data1.values)
        else:
            self.labels = self.clf.fit_predict(data1.values)
        self.output[self.output_variable] = xr.DataArray(self.labels, dims=[self.dim])
        self.output[self.output_variable].attrs.update(self.params)
        return self

    def _label(self, metric):
        """Perform the actual clustering.
        
        Parameters
        ----------
        metric : array-like
            The similarity matrix to cluster
        """
        self.clf.fit(metric)
        self.labels = self.clf.labels_


class GaussianMixtureModel(Labeler):
    """Gaussian Mixture Model for phase identification.
    
    Uses a Gaussian Mixture Model to identify phases based on their distribution
    in feature space. Particularly useful when phases are expected to have
    Gaussian distributions.

    Parameters
    ----------
    input_variable : str
        The name of the variable containing the feature data

    output_variable : str
        The name of the variable where labels will be stored

    dim : str
        The dimension name for samples in the dataset

    params : dict, optional
        Additional parameters for the GMM

    name : str, default='GaussianMixtureModel'
        The name to use when added to a Pipeline
    """

    def __init__(self, input_variable, output_variable, dim, params=None, name='GaussianMixtureModel'):
        super().__init__(name=name, input_variable=input_variable, output_variable=output_variable, dim=dim,
                         params=params)

    def _construct_labeler(self, **params):
        """Construct the GMM model.
        
        Parameters
        ----------
        **params : dict
            Additional parameters to update the GMM configuration
        """
        self.params.update(params)
        self.clf = sklearn.mixture.GaussianMixture(self.params['n_phases'])

    def calculate(self, dataset):
        """Apply GMM clustering to the dataset.
        
        Parameters
        ----------
        dataset : xr.Dataset
            The input dataset containing the feature data

        Returns
        -------
        Self
            The GaussianMixtureModel instance with computed labels
        """
        data1 = self._get_variable(dataset)
        self._construct_labeler()
        self.silhouette(data1.values)
        self.output[self.output_variable] = xr.DataArray(self.labels, dims=[self.dim])
        return self

    def _label(self, metric):
        """Perform the actual clustering.
        
        Parameters
        ----------
        metric : array-like
            The feature data to cluster
        """
        self.clf.fit(metric)
        self.labels = self.clf.predict(metric)


class AffinityPropagation(Labeler):
    """Affinity Propagation for phase identification.
    
    Uses Affinity Propagation clustering to identify phases. This method
    automatically determines the number of phases based on the data structure
    and similarity matrix.

    Parameters
    ----------
    input_variable : str
        The name of the variable containing the similarity matrix

    output_variable : str
        The name of the variable where labels will be stored

    dim : str
        The dimension name for samples in the dataset

    params : dict, optional
        Additional parameters for Affinity Propagation. Defaults include:
        - damping: 0.75
        - max_iter: 5000
        - convergence_iter: 250
        - affinity: 'precomputed'

    name : str, default='AffinityPropagation'
        The name to use when added to a Pipeline
    """

    def __init__(self, input_variable, output_variable, dim, params=None, name='AffinityPropagation'):
        super().__init__(name=name, input_variable=input_variable, output_variable=output_variable, dim=dim,
                         params=params)

        default_params = {}
        default_params['damping'] = 0.75
        default_params['max_iter'] = 5000
        default_params['convergence_iter'] = 250
        default_params['affinity'] = 'precomputed'
        for k,v in default_params.items():
            if k not in self.params:
                self.params[k] = v

    def calculate(self, dataset):
        """Apply Affinity Propagation clustering to the dataset.
        
        Parameters
        ----------
        dataset : xr.Dataset
            The input dataset containing the similarity matrix

        Returns
        -------
        Self
            The AffinityPropagation instance with computed labels
        """
        data1 = self._get_variable(dataset)
        
        self.clf = sklearn.cluster.AffinityPropagation(**self.params)
        self.clf.fit(data1.values)
        
        self.labels = self.clf.labels_
        self.output[self.output_variable] = xr.DataArray(self.labels, dims=[self.dim])
        self.output[self.output_variable].attrs.update(self.params)
        self.output[self.output_variable].attrs['n_phases'] = len(np.unique(self.labels))
        return self
        



import numpy as np
import gpflow
from gpflow.monitor import (
    ImageToTensorBoard,
    ModelToTensorBoard,
    Monitor,
    MonitorTaskGroup,
    ScalarToTensorBoard
)
import xarray as xr
from gpflow.optimizers import NaturalGradient
from gpflow import set_trainable
import tensorflow as tf
from scipy.stats import entropy
from sklearn.neighbors import KernelDensity
import tqdm
import warnings

from AFL.agent.util import ternary_to_xy

class HeteroscedasticGaussian(gpflow.likelihoods.Likelihood):
    """
    Defines the heteroscedastic gaussian likelihood 
    """
    def __init__(self, **kwargs):
        # this likelihood expects a single latent function F, and two columns in the data matrix Y:
        super().__init__(latent_dim=1, observation_dim=2, **kwargs)

    def _log_prob(self, F, Y):
        # log_prob is used by the quadrature fallback of variational_expectations and predict_log_density.
        # Because variational_expectations is implemented analytically below, this is not actually needed,
        # but is included for pedagogical purposes.
        # Note that currently relying on the quadrature would fail due to https://github.com/GPflow/GPflow/issues/966
        Y, NoiseVar = Y[:, 0], Y[:, 1]
        return gpflow.logdensities.gaussian(Y, F, NoiseVar)

    def _variational_expectations(self, Fmu, Fvar, Y):
        Y, NoiseVar = Y[:, 0], Y[:, 1]
        Fmu, Fvar = Fmu[:, 0], Fvar[:, 0]
        return (
            -0.5 * np.log(2 * np.pi)
            - 0.5 * tf.math.log(NoiseVar)
            - 0.5 * (tf.math.square(Y - Fmu) + Fvar) / NoiseVar
        )

    # The following two methods are abstract in the base class.
    # They need to be implemented even if not used.

    def _predict_log_density(self, Fmu, Fvar, Y):
        raise NotImplementedError

    def _predict_mean_and_var(self, Fmu, Fvar):
        raise NotImplementedError


class DummyGP:
    def __init__(self,dataset,kernel=None):
        self.dataset=dataset
        self.kernel=kernel
        
    def predict(self,components):
        domain = self.dataset.afl.comp.get(components)
        self.y_mean = np.zeros_like(domain)
        self.y_var = np.zeros_like(domain)
        return {'mean':self.y_mean,'var':self.y_var}

    
class GPR:
#   '''
#   This is a class for doing regression. 
#   -------------
#   The data construct needs to point to the values being regressed over, Y_i
#   If uncertainties are known for each value, set the heteroscedastic flag to True
#   
#   Pass the inputs and uncertainties into the initialization
#    '''    
#
    def __init__(self,dataset,inputs=None,uncertainties=None,kernel=None,heteroscedastic=False):
        self.kernel = kernel
        self.heteroscedastic = heteroscedastic
        self.reset_GP(dataset=dataset,kernel=kernel,heteroscedastic=heteroscedastic)
        self.iter_monitor = lambda x: None
        self.final_monitor = lambda x: None
        
    def construct_data(self,heteroscedastic=True, preprocess=True):
           
        domain = self.transform_domain()
        if heteroscedastic:
            inputs = self.dataset[self.dataset.attrs['AL_regression_data']]
            uncertainties = self.dataset[self.dataset.attrs['AL_regression_uncertainties']]
            targets = np.stack((inputs,uncertainties), axis=1)
            targets[:,0] = (targets[:,0] - targets[:,0].mean())/targets[:,0].std()
            targets[:,1] = targets[:,1]/targets[:,1].std()

        else:
            #targets = self.dataset(self.inputs)
            targets = xr.DataArray(np.expand_dims(np.array(self.dataset[self.dataset.attrs['AL_regression_data']].values),axis=1))
            targets = (targets - targets.mean())/targets.std()
        data = (domain,targets)
        return data
        

   # def transform_domain(self,components=None):
   ##     """
   ##     Transforms the evaluated space (i.e. compositions axes) to a range from 0-1. 
   ##     --------
   ##     """    
   #     if 'GP_domain_transform' in self.dataset.attrs:
   #         if self.dataset.attrs['GP_domain_transform']=='ternary':
   #             if not (len(self.dataset.attrs['components'])==3):   
   #                 raise ValueError("Ternary domain transform specified but len(components)!=3") 
   #             domain = self.dataset.afl.comp.ternary_to_xy(components=components)
   #         elif self.dataset.attrs['GP_domain_transform']=='standard_scaled':
   #             domain = self.dataset.afl.comp.get_standard_scaled(components=components)
   #         elif self.dataset.attrs['GP_domain_transform']=='range_scaled':
   #             components = self.dataset.afl.comp._get_default(components)
   #             ranges = {}
   #             for component in components:
   #                 ranges[component] = self.dataset.attrs[component+'_range'][1] - self.dataset.attrs[component+'_range'][0]
   #             domain = self.dataset.afl.comp.get_range_scaled(ranges=ranges,components=components)
   #         else:
   #             raise ValueError('Domain not recognized!')
   #     else:
   #         domain = self.dataset.afl.comp.get(components=components)
   #     return domain
            
    def transform_domain(self,components=None):
            
        if 'GP_domain_transform' in self.dataset.attrs:
            if self.dataset.attrs['GP_domain_transform']=='ternary':
                if not (len(self.dataset.attrs['components'])==3):   
                    raise ValueError("Ternary domain transform specified but len(components)!=3") 
                domain = self.dataset.afl.comp.ternary_to_xy(components=components)
            elif self.dataset.attrs['GP_domain_transform']=='standard_scaled':
                domain = self.dataset.afl.comp.get_standard_scaled(components=components)
            elif self.dataset.attrs['GP_domain_transform']=='range_scaled':
                components = self.dataset.afl.comp._get_default(components)
                ranges = {}
                for component in components:
                    ranges[component] = {}
                    ranges[component]['min'] = self.dataset.attrs[component+'_range'][0]
                    ranges[component]['max'] = self.dataset.attrs[component+'_range'][1]
                    ranges[component]['range'] = self.dataset.attrs[component+'_range'][1] - self.dataset.attrs[component+'_range'][0]
                domain = self.dataset.afl.comp.get_range_scaled(ranges,components=components)
            else:
                raise ValueError('Domain not recognized!')
        else:
            domain = self.dataset.afl.comp.get(components=components)
        return domain

    def reset_GP(self,dataset,kernel=None,heteroscedastic=False):
    #    """
    #   Constructs the GP model given a likelihood, the input data, a kernel function, and establishes an optimizer
    #   -------------
    #   
    #   """
        self.dataset = dataset
        data = self.construct_data(heteroscedastic)
            
        if kernel is None:
            kernel = gpflow.kernels.Matern32(variance=0.1,lengthscales=0.1) 
        
        if heteroscedastic:
            likelihood = HeteroskedasticGaussian()
            
            #instantiate the model with the data, kernel, and likelihood
            self.model = gpflow.models.VGP(
                data=data, 
                kernel=kernel, 
                likelihood=likelihood, 
                num_latent_gps=1
                )

            #This is from the aforementioned example. 
            #the adam optimizer algorithm is already implemented
            #the natgrad optimizer is necessary for the heteroscedastic regression, gamma is a fixed parameter here. may need to be optimized
            self.natgrad = NaturalGradient(gamma=0.5) 
            self.adam = tf.optimizers.legacy.Adam()
            set_trainable(self.model.q_mu, False)
            set_trainable(self.model.q_sqrt, False)
            
        else:
             self.model = gpflow.models.GPR(
                     data = data,
                     kernel = kernel,
                     )
             self.optimizer = tf.optimizers.legacy.Adam(learning_rate=0.001)
            

    def reset_monitoring(self,log_dir='test/',iter_period=1):
        model_task = ModelToTensorBoard(log_dir, self.model,keywords_to_monitor=['*'])
        lml_task   = ScalarToTensorBoard(log_dir, lambda: self.loss(), "Training Loss")
        
        fast_tasks = MonitorTaskGroup([model_task,lml_task],period=iter_period)
        self.iter_monitor = Monitor(fast_tasks)
        
        image_task = ImageToTensorBoard(
            log_dir, 
            self.plot, 
            "Mean/Variance",
            fig_kw=dict(figsize=(18,6)),
            subplots_kw=dict(nrows=1,ncols=3)
        )
        slow_tasks = MonitorTaskGroup(image_task) 
        self.final_monitor = Monitor(slow_tasks)

    def optimize(self,N=1000, gamma=0.5, final_monitor_step=None,progress_bar=False, tol=1e-4):
        i  = 0
        break_criteria = False
        
        if progress_bar:
            while (i<N) or (break_criteria==True):
                
                pre_step_HPs = np.array([i.numpy() for i in self.model.parameters])
                self._step(i)
                post_step_HPs = np.array([i.numpy() for i in self.model.parameters])
                i+=1
                if all(abs(pre_step_HPs-post_step_HPs) <= tol):
                    break_criteria=True
                    break
            # for i in tqdm.tqdm(tf.range(N),total=N):
            #     self._step(i)
        else:
            while (i<=N) or (break_criteria==True):

                pre_step_HPs = np.array([i.numpy() for i in self.model.parameters])
                self._step(i)
                post_step_HPs = np.array([i.numpy() for i in self.model.parameters])
                i+=1
                if all(abs(pre_step_HPs-post_step_HPs) <= tol):
                    break_criteria=True
                    break
            
        if final_monitor_step is None:
            final_monitor_step = i
        self.final_monitor(final_monitor_step)
           
    def _step(self,i):
        #the optimizers in the heteroscedastic GPFlow example

        if self.heteroscedastic:

            self.natgrad.minimize(self.model.training_loss, [(self.model.q_mu, self.model.q_sqrt)])
            self.adam.minimize(self.model.training_loss, self.model.trainable_variables)
        else:
            self.optimizer.minimize(self.model.training_loss, self.model.trainable_variables)
        self.iter_monitor(i)
    
    def predict(self,components):
        domain = self.transform_domain(components=components)
        self.y = self.model.predict_f(domain) 
        self.y_mean = self.y[0].numpy() 
        self.y_var = self.y[1].numpy() 
        return {'mean':self.y_mean,'var':self.y_var}

    



import requests
import base64
import pickle
from AFL.automation.APIServer.Client import Client
from AFL.automation.shared import serialization
import time

class AgentClient(Client):
    '''Communicate with NistoRoboto Agent server 

    '''
    
    def set_mask(self,mask):
        json = {}
        json['task_name'] = 'set_mask'
        json['mask'] = serialize(mask)
        json['serialized'] = True 
        self.enqueue(**json)
        
    def append_data(self,compositions,measurements,labels):
        json = {}
        json['task_name'] = 'append_data'
        json['compositions'] = serialize(compositions)
        json['measurements'] = serialize(measurements)
        json['labels'] = serialize(labels)
        self.enqueue(**json)
        
    def get_phasemap(self):
        json = {}
        json['task_name']  = 'get_object'
        json['name']  = 'phasemap'
        json['interactive']  = True
        retval = self.enqueue(**json)
        phasemap= deserialize(retval['return_val'])
        return phasemap

    # def get(self,name):
    #     json = {}
    #     json['task_name']  = 'get_object'
    #     json['name']  = name
    #     json['interactive']  = True
    #     retval = self.enqueue(**json)
    #     obj = deserialize(retval['return_val'])
    #     return obj
    # 
    # def set(self,**kw):
    #     json = {}
    #     json['task_name'] = 'set_object'
    #     for k,v in kw.items():
    #         json[k] = serialize(v)
    #     self.enqueue(**json)
    
    def _get_next_sample(self):
        response = requests.get(self.url+'/get_next_sample',headers=self.headers)
        if response.status_code != 200:
            raise RuntimeError(f'API call to _get_next_sample command failed with status_code {response.status_code}\n{response.text}')
        return deserialize(response.json())

    def get_next_sample_queued(self):
        json = {}
        json['task_name'] = 'get_next_sample'
        json['interactive'] = True
        retval = self.enqueue(**json)
        obj = deserialize(retval['return_val'])
        return obj[0]
    
    def get_next_sample(self,wait_on_stale=True):
        json = {}
        json['task_name']  = 'get_next'
        json['interactive']  = True
        while True:
            next_sample,stale = self._get_next_sample()
            if (not wait_on_stale) or (not stale):
                break
            else:
                time.sleep(2)
            
        return next_sample
    

#check if validated, if not error
    
    def predict(self):
        kw = {}
        kw['task_name'] = 'predict'
        self.enqueue(**kw)



import numpy as np
import copy
import scipy.spatial

import sklearn.mixture
import sklearn.cluster
from sklearn.metrics import pairwise

from scipy.spatial.distance import pdist, squareform
from scipy.spatial import distance as dist
from collections import Counter

import warnings 
from sklearn.metrics import silhouette_samples, silhouette_score
from collections import defaultdict

try:
    import tensorflow as tf
    import keras.models 
except ImportError:
    warnings.warn('Keras not installed. KerasClassifier will not work!')



class PhaseLabeler:
    def __init__(self,params=None):
        self.labels = None
        if params is None:
            self.params = {'n_phases':2}
        else:
            self.params = params
            
    def copy(self):
        return copy.deepcopy(self)
        
    def __getitem__(self,index):
        return self.labels[index]
    
    def __array__(self,dtype=None):
        return np.array(self.labels).astype(dtype)

    def remap_labels_by_count(self):
        label_map ={}
        for new_label,(old_label,_) in enumerate(sorted(Counter(self.labels).items(),key=lambda x: x[1],reverse=True)):
            label_map[old_label]=new_label
        self.labels = list(map(label_map.get,self.labels))
        
    def label(self):
        raise NotImplementedError('Sub-classes must implement label!')
        
    def silhouette(self,W):
        X = W.copy()
        silh_dict = defaultdict(list)
        max_n = min(X.shape[0],15)
        for n_phases in range(2,max_n):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._init(n_phases=n_phases)
                self._label(X)
            self.remap_labels_by_count()
            
            if len(np.unique(self.labels))==1:
                silh_scores = np.zeros(len(X))
            else:
                silh_scores = silhouette_samples(
                    1.0-X,
                    self,
                    metric='precomputed'
                )
            silh_dict['all_scores'].append(silh_scores)
            silh_dict['avg_scores'].append(silh_scores.mean())
            silh_dict['n_phases'].append(n_phases)
            silh_dict['labels'].append(self.labels.copy())
            # silh_dict['labelers'].append(self.copy())
    
        silh_avg = np.array(silh_dict['avg_scores'])
        found = False
        for cutoff in np.arange(0.85,0.4,-0.05):
            idx = np.where(silh_avg>cutoff)[0]
            if idx.shape[0]>0:
                idx = idx[-1]
                found=True
                break
                
        if not found:
            self.n_phases = 1
            self.labels=np.zeros(X.shape[0])
        else:
            self.n_phases = silh_dict['n_phases'][idx]
            self.labels = silh_dict['labels'][idx]
        self.silh_dict = silh_dict
        
class GaussianMixtureModel(PhaseLabeler):
    def __init__(self,params=None):
        super().__init__(params)
        self.name = f'GaussianMixtureModel'
        
    def label(self,phasemap,**params):
        self._init(**params)
        self.silhouette(phasemap.W.values)
        
    def _init(self,**params):
        if params:
            self.params.update(params)
        self.clf = sklearn.mixture.GaussianMixture(self.params['n_phases'])
        
    def _label(self,metric):
        self.clf.fit(metric)
        self.labels = self.clf.predict(metric)
        
class SpectralClustering(PhaseLabeler):
    def __init__(self,params=None):
        super().__init__(params)
        self.name = f'SpectralClustering'
        
    def label(self,phasemap,**params):
        self._init(**params)
        self.silhouette(phasemap.W.values)
        
    def _init(self,**params):
        if params:
            self.params.update(params)
            
        self.clf = sklearn.cluster.SpectralClustering(
            self.params['n_phases'],
            affinity = 'precomputed',  
            assign_labels="discretize",  
            random_state=0,  
            n_init = 1000
        )
        
    def _label(self,metric):
        self.clf.fit(metric)
        self.labels = self.clf.labels_

        
    
class DBSCAN(PhaseLabeler):#!!!
    def __init__(self,params=None):
        super().__init__(params)
        self.name = f'DBSCAN'
        if 'eps' not in self.params:
            self.params['eps'] = 0.05
        
    def label(self,phasemap,**params):
        if params:
            self.params.update(params)
            
        self.clf = sklearn.cluster.DBSCAN(
            eps=self.params['eps'],
            metric='precomputed',
            min_samples=1
        )
        #need something that works like distance, not similarty
        self.clf.fit(1.0-phasemap.W.values)
        self.labels = self.clf.labels_
        self.n_phases = len(np.unique(self.labels))
        return self.labels

class KerasClassifier(PhaseLabeler):
    def __init__(self,model_path,model_q,params=None,data_variable='deriv0'):
        super().__init__(params)
        self.name = f'KerasClassifier'

        self.model_path = model_path
        self.model_q = model_q
        self.clf = None
        self.data_variable = data_variable
        
    def label(self,phasemap,transpose_var='sample',**params):
        if self.clf is None:
            with tf.device('/CPU:0'):
                self.clf = keras.models.load_model(self.model_path)
        X_data = phasemap[self.data_variable].transpose(transpose_var,...).interp(logq=np.log10(self.model_q)).values
        self.labels = self.clf.predict(X_data).argmax(axis=1)
        self.n_phases = len(np.unique(self.labels))

class AffinityPropagation(PhaseLabeler):
    def __init__(self,params=None):
        super().__init__(params)
        self.name = f'AffinityPropagation'
        self.params['damping'] = 0.75
        self.params['max_iter'] = 5000
        self.params['convergence_iter'] = 250
        self.params['affinity'] = 'precomputed'
        if 'n_phases' in self.params:
            del self.params['n_phases']
        if params is not None:
            self.params.update(params)
        
    def label(self,phasemap,**params):
        self._init(**params)
        self.clf.fit(phasemap.W.values)
        self.labels = self.clf.labels_
        self.n_phases = len(np.unique(self.labels))
        
    def _init(self,**params):
        if params:
            self.params.update(params)
            
        self.clf = sklearn.cluster.AffinityPropagation(**self.params)



from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
import time
import logging
import datetime

class FileChangeHandler(PatternMatchingEventHandler):
    def __init__(self,fname,callback,cooldown,logger):
        super().__init__(patterns=[f'*/{fname}'])
        self.callback = callback
        self.cooldown = datetime.timedelta(seconds=cooldown)
        self.last_call = datetime.datetime.now()-self.cooldown #make sure first call happens
        self.logger = logger
        
    def on_modified(self, event):
        if (datetime.datetime.now()-self.last_call)>=self.cooldown:
            self.logger.info(f'Watchdog triggered with event: {event}')
            self.last_call = datetime.datetime.now()
            self.callback()

class WatchDog:
    def __init__(self,path,fname,callback,cooldown):
        self.logger = logging.getLogger()
        self.event_handler = FileChangeHandler(fname=fname,callback=callback,cooldown=cooldown,logger=self.logger)
        self.observer = Observer()
        self.path = path
        self.fname = fname
        
    def start(self):
        self.observer.schedule(self.event_handler, path=self.path)
        self.logger.info('Starting WatchDog on dir {self.path} and file {self.fname}')
        self.observer.start()
        
    def stop(self):
        self.observer.stop()
        self.observer.join()
        self.logger.info('Stopping WatchDog')
        



from AFL.automation.APIServer.Client import Client
from AFL.automation.prepare.OT2Client import OT2Client
from AFL.automation.shared.utilities import listify
from AFL.automation.APIServer.Driver import Driver
from AFL.automation.shared import serialization
from AFL.automation.shared.utilities import mpl_plot_to_bytes
from AFL.automation.shared.utilities import listify

from math import ceil,sqrt
import json
import time
import requests
import shutil
import datetime
import traceback
import warnings

import pandas as pd
import numpy as np
import xarray as xr
import pathlib
import tqdm

import h5py

import io
import matplotlib.pyplot as plt
import matplotlib

import AFL.agent.xarray_extensions
from AFL.agent import AcquisitionFunction 
from AFL.agent import GaussianProcess 
from AFL.agent import HscedGaussianProcess 
from AFL.agent import Metric 
from AFL.agent import PhaseLabeler
from AFL.agent.WatchDog import WatchDog


import tensorflow as tf

import gpflow

import uuid

class Virtual_Multimodal_AgentDriver(Driver):
    defaults={}
    defaults['compute_device'] = '/device:CPU:0'
    defaults['data_path'] = '~/'
    defaults['AL_manifest_file'] = 'manifest.csv'
    defaults['save_path'] = '/home/AFL/'
    defaults['data_tag'] = 'default'
    defaults['subtract_background'] = False
    # defaults['grid_pts_per_row'] = 100
    def __init__(self,overrides=None):
        Driver.__init__(self,name='SAS_AgentDriver',defaults=self.gather_defaults(),overrides=overrides)

        self.watchdog = None 
        self.AL_manifest = None
        self._app = None
        self.name = 'SAS_AgentDriver'

        self.status_str = 'Fresh Server!'

        self.dataset = None
        self.n_phases = None
        self.metric = Metric.Dummy()
        self.acquisition = None
        self.labeler = None
        self.next_sample = None
        self._mask = None
        self.iteration = 0
        self.acq_count = 0
        self.sample_randomly = False
    @property
    def app(self):
        return self._app
    
    @app.setter
    def app(self,value):
        self._app = value
        # if value is not None:
        #     self.reset_watchdog()

    def status(self):
        status = []
        status.append(self.status_str)
        if self.metric is not None:
            status.append(f'Metric: {self.metric.name}')
        else:
            status.append(f'Metric: No metric loaded')
        if self.acquisition is not None:
            status.append(f'Acq.: {self.acquisition.name}')
        else:
            status.append(f'Acq.: No acquisition function loaded')
        if self.labeler is not None:
            status.append(f'Labeler: {self.labeler.name}')
        else:
            status.append(f'Labeler: No labeler loaded')
        #if self.mask is not None:
        #    status.append(f'Masking: {self.masked_points}/{self.total_points}')
        #else:
        #    status.append(f'Masking: No mask loaded')
        if self.n_phases is not None:
            status.append(f'Found {self.n_phases} phases')
            
        status.append(f'Using {self.config["compute_device"]}')
        status.append(f'Data Manifest:{self.config["AL_manifest_file"]}')
        status.append(f'Iteration {self.iteration}')
        status.append(f'Acquisition Count {self.acq_count}')
        return status
    
    def reset_watchdog(self):
        if not (self.watchdog is None):
            self.watchdog.stop()
            
        if self.app is not None:
            logger = self.app.logger
        else:
            logger = None
        
        path = pathlib.Path(self.config['AL_manifest_file'])
        self.watchdog = WatchDog(
            path=path.parent,
            fname=path.name,
            callback=self.update_phasemap,
            cooldown=5,
        )
        self.watchdog.start()
        
    def update_status(self,value):
        self.status_str = value
        self.app.logger.info(value)
    
    @property
    def mask(self):
        return self._mask
    
    @mask.setter
    def mask(self,value):
        self._mask=value
    
    def subtract_background(self,path,fname):

        raw = self.read_csv(path,fname)
        raw.name = 'raw'

        empty_fname = 'MT-'+str(fname)
        empty = self.read_csv(path,empty_fname)
        empty.name = 'empty'

        sample_name = fname.replace('_chosen_r1d.csv','')
        h5_path = path / (sample_name+'.h5')
        with h5py.File(h5_path,'r') as h5:
            sample_transmission = h5['entry/sample/transmission'][()]

        empty_name = empty_fname.replace('_chosen_r1d.csv','')
        h5_path = path / (empty_name+'.h5')
        with h5py.File(h5_path,'r') as h5:
            empty_transmission = h5['entry/sample/transmission'][()]

        corrected = raw#/sample_transmission - empty.interp_like(raw)/empty_transmission
        corrected.name = 'corrected'
        return corrected,raw,empty,sample_transmission,empty_transmission

        
    def read_data(self):
        '''Read and process a dataset from a netcdf file'''
        self.update_status(f'Reading the latest data in {self.config["AL_manifest_file"]}')
        path = pathlib.Path(self.config['data_path'])
        
        self.dataset = xr.load_dataset(path / self.config['AL_manifest_file'])

        if 'savgol_filter' in self.dataset.attrs:
            for AL_data in listify(self.dataset.attrs['savgol_filter']):
                measurement = self.dataset[AL_data]
                kwargs = {
                    "xname":self.dataset.attrs.get(AL_data+'_savgol_xname','q'),
                    "yname":self.dataset.attrs.get(AL_data+'_savgol_yname','SAS'),
                    "xlo":self.dataset.attrs.get(AL_data+'_savgol_xlo',None),
                    "xhi":self.dataset.attrs.get(AL_data+'_savgol_xhi',None),
                    "xlo_isel":self.dataset.attrs.get(AL_data+'_savgol_xlo_isel',None),
                    "xhi_isel":self.dataset.attrs.get(AL_data+'_savgol_xhi_isel',None),
                    "pedestal":self.dataset.attrs.get(AL_data+'_savgol_pedestal',None),
                    "npts":self.dataset.attrs.get(AL_data+'_savgol_npts',None),
                    "sgf_window_length":self.dataset.attrs.get(AL_data+'_savgol_window_length',3),
                    "sgf_polyorder":self.dataset.attrs.get(AL_data+'_savgol_polyorder',2),
                    "logx":self.dataset.attrs.get(AL_data+'_savgol_logx',False),
                    "logy":self.dataset.attrs.get(AL_data+'_savgol_logy',False),
                }
                self.dataset[AL_data+'_deriv0'] = measurement.afl.util.preprocess_data(derivative=0,**kwargs)
                self.dataset[AL_data+'_deriv1'] = measurement.afl.util.preprocess_data(derivative=1,**kwargs)
                self.dataset[AL_data+'_deriv2'] = measurement.afl.util.preprocess_data(derivative=2,**kwargs)

                self.dataset[AL_data+'_deriv0'] = self.dataset[AL_data+'_deriv0'] - self.dataset[AL_data+'_deriv0'].min('logq')
        
        if 'labels' not in self.dataset:
            self.dataset = self.dataset.afl.labels.make_default()

        if self.mask is not None:
            if 'mask' in self.dataset:
                raise ValueError('Both AgentServer and phasemap from netcdf have mask!')
            self.dataset.update(self.mask)

        self.components = self.dataset.attrs['components']

    def label(self):

        self.update_status(f'Labeling data on iteration {self.iteration}')

        self.metric.calculate(self.dataset)
        self.dataset['W'] = (('sample_i','sample_j'),self.metric.W)
        self.dataset.attrs['metric'] = str(self.metric.to_dict())
        
        self.labeler.label(self.dataset)
        self.n_phases = self.labeler.n_phases

        self.update_status(f'Found {self.labeler.n_phases} phases')

        self.dataset.attrs['n_phases'] = self.n_phases
        self.dataset['labels'] = ('sample',self.labeler.labels)
        self.dataset = self.dataset.afl.labels.make_ordinal()
        
    def extrapolate(self):
        # Predict phase behavior at each point in the phase diagram
        if self.n_phases==1:
            self.update_status(f'Using dummy GP for one phase')
            self.classifier_GP = GaussianProcess.DummyGP(self.dataset)
        else:
            with tf.device(self.config['compute_device']):
                #the defalut AL mode is multimodal_similarity
                self.update_status(f'Starting gaussian process classifier calculation on {self.config["compute_device"]}')
                kernel_classifier = gpflow.kernels.Matern32(variance=0.5,lengthscales=1.0) 
                self.classifier_GP = GaussianProcess.GP(
                    dataset = self.dataset,
                    kernel  = kernel_classifier
                )
                self.classifier_GP.optimize(2000,progress_bar=True)
                

                if self.dataset.attrs['AL_mode'] == 'bimodal_UCB':  
                    kernel_regressor =  gpflow.kernels.Matern52(variance=0.5,lengthscales=2.0) + gpflow.kernels.White(1e-1)
                    
                    if self.dataset.attrs['regression_mode'] == 'heteroscedastic':
                        self.update_status(f'Starting heteroscedastic gaussian process regression calculation on {self.config["compute_device"]}')

                        self.regressor_GP = HscedGaussianProcess.GPR(
                           dataset = self.dataset,
                           kernel  = kernel_regressor
                           )


                    elif self.dataset.attrs['regression_mode'] == 'homoscedastic':
                        self.update_status(f'Starting homoscedastic gaussian process regression calculation on {self.config["compute_device"]}')
                        self.regressor_GP = HscedGaussianProcess.GPR(
                            dataset = self.dataset,
                            kernel = kernel_regressor,
                            heteroscedastic = False
                            )

                    
                    self.regressor_GP.optimize(500,progress_bar=True)

                
        self.acq_count   = 0
        self.iteration  += 1 #iteration represents number of full calculations

        self.update_status(f'Finished AL iteration {self.iteration}')
        
    @Driver.unqueued()
    def get_next_sample(self):
        self.update_status(f'Calculating acquisition function...')
        self.acquisition.reset_phasemap(self.dataset)
        if self.dataset.attrs['AL_mode'] == 'multimodal_similarity':
            self.dataset = self.acquisition.calculate_metric(self.classifier_GP)

        elif self.dataset.attrs['AL_mode'] == 'bimodal_UCB':
            self.dataset = self.acquisition.calculate_metric(
                    GP  = self.classifier_GP,
                    GPR = self.regressor_GP,
                    )

        self.update_status(f'Finding next sample composition based on acquisition function')
        composition_check = self.dataset[self.dataset.attrs['components']]
        if 'sample' in composition_check.indexes:
            composition_check = composition_check.reset_index('sample').reset_coords(drop=True)
        composition_check = composition_check.to_array('component').transpose('sample',...)
        self.next_sample = self.acquisition.get_next_sample(
            composition_check = composition_check,
            sample_randomly=self.sample_randomly
        )
        
        next_dict = self.next_sample.squeeze().to_pandas().to_dict()
        status_str = 'Next sample is '
        for k,v in next_dict.items():
            status_str += f'{k}: {v:4.3f} '
        self.update_status(status_str.strip())
        self.acq_count+=1#acq represents number of times 'get_next_sample' is called
        
    @Driver.unqueued()
    def save_results(self,sample_uuid):
        #write netcdf
        uuid_str = str(uuid.uuid4())
        save_path = pathlib.Path(self.config['save_path'])
        date =  datetime.datetime.now().strftime('%y%m%d')
        time =  datetime.datetime.now().strftime('%H:%M:%S')
        self.dataset['gp_classifier_y_var'] = (('grid','phase_num'),self.acquisition.y_var_GPC)
        self.dataset['gp_classifier_y_mean'] = (('grid','phase_num'),self.acquisition.y_mean_GPC)
        if self.dataset.attrs['AL_mode'] == 'bimodal_UCB':
            self.dataset['gp_regressor_y_mean'] = (('grid',self.acquisition.y_mean_GPR.squeeze()))
            self.dataset['gp_regressor_y_var']  = (('grid',self.acquisition.y_var_GPR.squeeze()))
            if self.acquisition.regressor_sample is not None:
                self.dataset['gp_regressor_y_mean_sample']  = (('grid',self.acquisition.regressor_sample))

        #self.dataset['next_sample'] = ('component',self.next_sample.squeeze().values)
        #reset_index('grid').drop(['SLES3_grid','DEX_grid','CAPB_grid'])
        
        from xarray.core.merge import MergeError
        
        if 'component' in self.dataset.dims:
            warnings.warn('Dropping component dim (and all associated vars) from phasemap...')
            self.dataset = self.dataset.drop_dims('component')
            
        try:
            self.dataset['next_sample'] = self.acquisition.next_sample.squeeze()
        except MergeError:
            self.dataset['next_sample'] = self.acquisition.next_sample.squeeze().reset_coords(drop=True)

        if 'mask' in self.dataset:
            self.dataset['mask'] = self.dataset['mask'].astype(int)
            
        self.dataset.attrs['uuid'] = uuid_str
        self.dataset.attrs['sample_uuid'] = sample_uuid
        self.dataset.attrs['date'] = date
        self.dataset.attrs['time'] = time
        self.dataset.attrs['data_tag'] = self.config["data_tag"]
        self.dataset.attrs['acq_count'] = self.acq_count
        self.dataset.attrs['iteration'] = self.iteration
        self.dataset.to_netcdf(save_path/f'phasemap_{self.config["data_tag"]}_{uuid_str}.nc')
        
        #write manifest csv
        AL_manifest_path = save_path/'calculation_manifest.csv'
        if AL_manifest_path.exists():
            self.AL_manifest = pd.read_csv(AL_manifest_path)
        else:
            self.AL_manifest = pd.DataFrame(columns=['uuid','date','time','data_tag','iteration','acq_count'])

        
        row = {}
        row['uuid'] = uuid_str
        row['sample_uuid'] = sample_uuid
        row['date'] =  date
        row['time'] =  time
        row['data_tag'] = self.config['data_tag']
        row['iteration'] = self.iteration
        row['acq_count'] = self.acq_count
        self.AL_manifest = pd.concat([self.AL_manifest.T,pd.Series(row)],axis=1,ignore_index=True).T
        self.AL_manifest.to_csv(AL_manifest_path,index=False)

    def predict(self,datatype=None,sample_uuid=None):
        self.read_data()
        self.label()
        self.extrapolate()
        self.get_next_sample()
        self.save_results(sample_uuid)

    @Driver.unqueued(render_hint='precomposed_svg')
    def plot_scatt(self,**kwargs):
        if self.dataset is not None:
            if 'labels_ordinal' not in self.dataset:
                self.dataset['labels_ordinal'] = ('system',np.zeros(self.dataset.sizes['sample']))
                labels = [0]
            else:
                labels = np.unique(self.dataset.labels_ordinal.values)
            
        if 'precomposed' in kwargs['render_hint']:
            matplotlib.use('Agg') #very important
            if self.dataset is None:
                fig,ax = plt.subplots()
                plt.text(1,5,'No phasemap loaded. Run .read_data()')
                plt.gca().set(xlim=(0,10),ylim=(0,10))
            else:
                N = len(labels)
                fig,axes = plt.subplots(N,2,figsize=(8,N*4))

                if N==1:
                    axes = np.array([axes])

                for i,label in enumerate(labels):
                    spm = self.dataset.set_index(sample='labels_ordinal').sel(sample=label)
                    plt.sca(axes[i,0])
                    spm.data.afl.scatt.plot_linlin(x='logq',legend=False);
                
                    plt.sca(axes[i,1])
                    spm.afl.comp.plot_discrete(components=self.dataset.attrs['components'][:3]);
    
            svg  = mpl_plot_to_bytes(fig,format='svg')
            return svg
        elif kwargs['render_hint']=='raw': 
            # construct dict to send as json (all np.ndarrays must be converted to list!)
            out_dict = {}
            if self.dataset is None:
                out_dict = {'Error': 'No phasemap loaded. Run .read_data()'}
            else:
                out_dict['components'] = self.dataset.attrs['components']
                for i,label in enumerate(labels):
                    out_dict[f'phase_{i}'] = {}
                    
                    spm = self.dataset.set_index(sample='labels_ordinal').sel(sample=label)
                    out_dict[f'phase_{i}']['labels'] = list(spm.labels.values)
                    out_dict[f'phase_{i}']['labels_ordinal'] = int(label)
                    out_dict[f'phase_{i}']['q'] = list(spm.q.values)
                    #out_dict[f'phase_{i}']['raw_data'] = list(spm.raw_data.values)
                    out_dict[f'phase_{i}']['data'] = list(spm.data.values)
                    out_dict[f'phase_{i}']['compositions'] = {}
                    for component in self.dataset.attrs['components']:
                        out_dict[f'phase_{i}']['compositions'][component] = list(spm[component].values)
            return out_dict
        else:
            raise ValueError(f'Cannot handle render_hint={kwargs["render_hint"]}')

    @Driver.unqueued(render_hint='precomposed_svg')
    def plot_acq(self,**kwargs):
        matplotlib.use('Agg') #very important
        fig,ax = plt.subplots()
        if self.dataset is None:
            plt.text(1,5,'No phasemap loaded. Run .read_data()')
            plt.gca().set(xlim=(0,10),ylim=(0,10))
        else:
            self.acquisition.plot()
        svg  = mpl_plot_to_bytes(fig,format='svg')
        return svg

    @Driver.unqueued(render_hint='precomposed_svg')
    def plot_gp(self,**kwargs):
        if self.dataset is None:
            return 'No phasemap loaded. Run read_data()'

        if 'gp_classifier_y_mean' not in self.dataset:
            raise ValueError('No GP results in phasemap. Run .predict()')

        if 'precomposed' in kwargs['render_hint']:
            matplotlib.use('Agg') #very important
            N = self.dataset.sizes['phase_num']
            fig,axes = plt.subplots(self.dataset.sizes['phase_num'],2,figsize=(8,N*4))
            if N==1:
                axes = np.array([axes])
            i = 0
            for (_,labels1),(_,labels2) in zip(self.dataset.gp_classifier_y_mean.groupby('phase_num'),self.dataset.gp_classifier_y_var.groupby('phase_num')):
                plt.sca(axes[i,0])
                self.dataset.where(self.dataset.mask).afl.comp.plot_continuous(components=self.dataset.attrs['components_grid'],cmap='magma',labels=labels1.values);
                plt.sca(axes[i,1])
                self.dataset.where(self.dataset.mask).afl.comp.plot_continuous(components=self.dataset.attrs['components_grid'],labels=labels2.values);
                i+=1

            img  = mpl_plot_to_bytes(fig,format=kwargs['render_hint'].split('_')[1])
            return img

        elif kwargs['render_hint']=='raw': 
            # construct dict to send as json (all np.ndarrays must be converted to list!)

            out_dict = {}
            out_dict['components'] = self.dataset.attrs['components']
            out_dict['components_grid'] = self.dataset.attrs['components_grid']
            for component in self.dataset.attrs['components_grid']:
                out_dict[component] = list(self.dataset[component].values)

            x,y = self.dataset.afl.comp.to_xy(self.dataset.attrs['components_grid']).T
            out_dict['x'] = list(x)
            out_dict['y'] = list(y)

            i =0
            for (_,labels1),(_,labels2) in zip(self.dataset.gp_classifier_y_mean.groupby('phase_num'),self.dataset.gp_classifier_y_var.groupby('phase_num')):
                out_dict[f'phase_{i}'] = {'mean':list(labels1.values),'var':list(labels2.values)}
                i+=1
            return out_dict
        else:
            raise ValueError(f'Cannot handle render_hint={kwargs["render_hint"]}')



    


   



"""PipelineOps for Pairwise Metrics

This module contains operations that compute pairwise relationships between samples.
PairMetrics generate matrices that capture similarity, distance, or other relationships between pairs of data points.

These metrics are useful for:
- Measuring similarity or distance between samples
- Constructing adjacency matrices for graph-based algorithms
- Identifying clusters or patterns in data
- Quantifying relationships between different observations

Each PairMetric is implemented as a PipelineOp that can be composed with others in a processing pipeline.
"""

import copy
from numbers import Number
from typing import Optional, List, Dict, Any

import numpy as np
import scipy.spatial  # type: ignore
import xarray as xr
from sklearn.metrics import pairwise  # type: ignore
from typing_extensions import Self

from AFL.double_agent.PipelineOp import PipelineOp
from AFL.double_agent.util import listify


class PairMetric(PipelineOp):
    """Base class for all PairMetrics that produce similarity or distance matrices

    This abstract base class provides common functionality for computing and manipulating
    pairwise metrics between samples. It handles similarity constraints, normalization,
    and provides a framework for different metric implementations.

    Parameters
    ----------
    input_variable : str | List[str]
        The name of the data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the dataset
    sample_dim : str, default="sample"
        The dimension containing different samples
    params : Optional[Dict[str, Any]], default=None
        Additional parameters for metric calculation
    constrain_same : Optional[List], default=None
        List of pairs that should have perfect similarity
    constrain_different : Optional[List], default=None
        List of pairs that should have zero similarity
    name : str, default="PairMetric"
        The name to use when added to a Pipeline
    """
    def __init__(
        self,
        input_variable: str | List[str],
        output_variable: str,
        sample_dim: str = "sample",
        params: Optional[Dict[str, Any]] = None,
        constrain_same: Optional[List] = None,
        constrain_different: Optional[List] = None,
        name: str = "PairMetric",
    ) -> None:

        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.W = np.eye(1)
        self.sample_dim = sample_dim

        if params is None:
            self.params = {}
        else:
            self.params = params

        if constrain_same is None:
            self.constrain_same = []
        else:
            self.constrain_same = constrain_same

        if constrain_different is None:
            self.constrain_different = []
        else:
            self.constrain_different = constrain_different

        self._banned_from_attrs.append("W")
        self._banned_from_attrs.append("params")
        self._banned_from_attrs.append("combine")

    def __getitem__(self, index):
        return self.W[index]

    def __array__(self, dtype=None):
        return self.W.astype(dtype)

    def normalize1(self) -> np.ndarray:
        """Normalize similarity matrix such that the diagonal values are all equal to 1"""
        W = self.W.copy()
        diag = np.diag(W)
        W = W / np.sqrt(np.outer(diag, diag))
        return W

    def normalize2(self):
        """Normalize similarity matrix such that all values are between 0 and 1"""
        W = self.W.copy()
        W = (W - W.min()) / (W.max() - W.min())
        return W

    def apply_constraints(self):
        """
        Constrain pairs in the similarity matrix to be perfectly similar (S[i,j]=1.0) or
        perfectly dissimilar (S[i,j]=0.0).
        """
        W = self.W.copy()
        for i, j in self.constrain_same:
            W[i, j] = 1.0
            W[j, i] = 1.0

        for i, j in self.constrain_different:
            W[i, j] = 0.0
            W[j, i] = 0.0
        return W

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.Dataset`"""
        return NotImplementedError(".calculate must be implemented in subclasses")  # type: ignore


class Dummy(PairMetric):
    """PairMetric that returns only self-similarity (identity matrix)
    
    This simple metric creates an identity matrix where diagonal elements (self-similarity)
    are 1.0 and all off-diagonal elements are 0.0. This can be useful as a baseline
    or for testing purposes.

    Parameters
    ----------
    input_variable : str
        The name of the data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the dataset
    sample_dim : str
        The dimension containing different samples
    params : Optional[Dict[str, Any]], default=None
        Additional parameters for metric calculation (not used in this class)
    constrain_same : Optional[List], default=None
        List of pairs that should have perfect similarity
    constrain_different : Optional[List], default=None
        List of pairs that should have zero similarity
    name : str, default="DummyMetric"
        The name to use when added to a Pipeline
    """
    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        params: Optional[Dict[str, Any]] = None,
        constrain_same: Optional[List] = None,
        constrain_different: Optional[List] = None,
        name: str = "DummyMetric",
    ) -> None:

        super().__init__(
            name=name,
            input_variable=input_variable,
            output_variable=output_variable,
            sample_dim=sample_dim,
            params=params,
            constrain_same=constrain_same,
            constrain_different=constrain_different,
        )

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.Dataset`"""
        n_samples = dataset.sizes[self.sample_dim]
        self.W = np.identity(n_samples)
        return self


class Similarity(PairMetric):
    """Computes pairwise similarity between samples using kernel functions
    
    This class uses scikit-learn's pairwise_kernels to compute similarity matrices
    between samples. Various kernel functions can be specified through the params
    dictionary (e.g., 'linear', 'rbf', 'polynomial').

    For details on available kernel functions and their parameters, see:
    https://scikit-learn.org/stable/modules/metrics.html#metrics

    Parameters
    ----------
    input_variable : str
        The name of the data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the dataset
    sample_dim : str
        The dimension containing different samples
    params : Optional[Dict[str, Any]], default=None
        Parameters for the kernel function, including 'metric' to specify the kernel type
    constrain_same : Optional[List], default=None
        List of pairs that should have perfect similarity
    constrain_different : Optional[List], default=None
        List of pairs that should have zero similarity
    name : str, default="SimilarityMetric"
        The name to use when added to a Pipeline
    """
    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        params: Optional[Dict[str, Any]] = None,
        constrain_same: Optional[List] = None,
        constrain_different: Optional[List] = None,
        name="SimilarityMetric",
    ) -> None:

        super().__init__(
            name=name,
            input_variable=input_variable,
            output_variable=output_variable,
            sample_dim=sample_dim,
            params=params,
            constrain_same=constrain_same,
            constrain_different=constrain_different,
        )

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.Dataset`"""

        data1 = self._get_variable(dataset)

        if len(data1.shape) == 1:
            data1 = data1.expand_dims("feature", axis=1)
        self.W = pairwise.pairwise_kernels(data1, filter_params=True, **self.params)

        dims = [self.sample_dim + "_i", self.sample_dim + "_j"]
        self.output[self.output_variable] = xr.DataArray(self.W, dims=dims)  # type: ignore
        self.output[self.output_variable].attrs.update(self.params)  # type: ignore

        return self


class Distance(PairMetric):
    """Computes pairwise distances between samples
    
    This class uses scikit-learn's pairwise_distances to compute distance matrices
    between samples. Various distance metrics can be specified through the params
    dictionary (e.g., 'euclidean', 'manhattan', 'cosine').

    For details on available distance metrics and their parameters, see:
    https://scikit-learn.org/stable/modules/metrics.html#metrics

    Parameters
    ----------
    input_variable : str
        The name of the data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the dataset
    sample_dim : str
        The dimension containing different samples
    params : Optional[Dict[str, Any]], default=None
        Parameters for the distance function, including 'metric' to specify the distance type
    constrain_same : Optional[List], default=None
        List of pairs that should have perfect similarity
    constrain_different : Optional[List], default=None
        List of pairs that should have zero similarity
    name : str, default="DistanceMetric"
        The name to use when added to a Pipeline
    """
    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        params: Optional[Dict[str, Any]] = None,
        constrain_same: Optional[List] = None,
        constrain_different: Optional[List] = None,
        name="DistanceMetric",
    ) -> None:

        super().__init__(
            name=name,
            input_variable=input_variable,
            output_variable=output_variable,
            sample_dim=sample_dim,
            params=params,
            constrain_same=constrain_same,
            constrain_different=constrain_different,
        )

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.Dataset`"""
        data1 = self._get_variable(dataset)

        self.W = pairwise.pairwise_distances(data1, **self.params)

        dims = [self.sample_dim + "_i", self.sample_dim + "_j"]
        self.output[self.output_variable] = xr.DataArray(self.W, dims=dims)  # type: ignore
        return self


class Delaunay(PairMetric):
    """Creates a similarity matrix based on Delaunay triangulation
    
    This class constructs a binary adjacency matrix where samples that share
    an edge in the Delaunay triangulation have a similarity of 1.0, and all
    other pairs have a similarity of 0.0. This is useful for identifying
    natural neighbors in the data.

    Parameters
    ----------
    input_variable : str
        The name of the data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the dataset
    sample_dim : str
        The dimension containing different samples
    params : Optional[Dict[str, Any]], default=None
        Additional parameters (not used in this class)
    constrain_same : Optional[List], default=None
        List of pairs that should have perfect similarity
    constrain_different : Optional[List], default=None
        List of pairs that should have zero similarity
    name : str, default="DelaunayMetric"
        The name to use when added to a Pipeline
    """
    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        params: Optional[Dict[str, Any]] = None,
        constrain_same: Optional[List] = None,
        constrain_different: Optional[List] = None,
        name="DelaunayMetric",
    ) -> None:

        super().__init__(
            name=name,
            input_variable=input_variable,
            output_variable=output_variable,
            sample_dim=sample_dim,
            params=params,
            constrain_same=constrain_same,
            constrain_different=constrain_different,
        )

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.Dataset`"""
        data1 = self._get_variable(dataset)
        tri = scipy.spatial.Delaunay(data1.values)
        edges_explicit = np.concatenate(
            (tri.vertices[:, :2], tri.vertices[:, 1:], tri.vertices[:, ::2]), axis=0
        )
        adj = np.zeros((data1.values.shape[0], data1.values.shape[0]))
        adj[edges_explicit[:, 0], edges_explicit[:, 1]] = 1.0
        self.W = np.clip(adj + adj.T, 0, 1)

        return self


class CombineMetric(PairMetric):
    """Combines multiple similarity/distance matrices into a single matrix
    
    This class allows for the combination of multiple similarity or distance matrices
    using either product or sum operations. Each matrix can be weighted differently
    using powers (for product) or coefficients (for sum).

    Parameters
    ----------
    input_variables : List[str]
        List of variable names containing similarity/distance matrices to combine
    output_variable : str
        The name of the variable to be inserted into the dataset
    sample_dim : str
        The dimension containing different samples
    combine_by : str
        Method to combine matrices, either "prod" (product) or "sum"
    combine_by_powers : Optional[List[Number]], default=None
        List of powers to apply to each matrix when using "prod" combination
    combine_by_coeffs : Optional[List[Number]], default=None
        List of coefficients to multiply each matrix by when using "sum" combination
    params : Optional[str], default=None
        Additional parameters
    constrain_same : Optional[List], default=None
        List of pairs that should have perfect similarity
    constrain_different : Optional[List], default=None
        List of pairs that should have zero similarity
    name : str, default="CombineMetric"
        The name to use when added to a Pipeline
    """
    def __init__(
        self,
        input_variables: List[str],
        output_variable: str,
        sample_dim: str,
        combine_by: str,
        combine_by_powers: Optional[List[Number]] = None,
        combine_by_coeffs: Optional[List[Number]] = None,
        params: Optional[str] = None,
        constrain_same: Optional[List] = None,
        constrain_different: Optional[List] = None,
        name="CombineMetric",
    ) -> None:

        super().__init__(
            name=name,
            input_variable=input_variables,
            output_variable=output_variable,
            sample_dim=sample_dim,
            params=params,
            constrain_same=constrain_same,
            constrain_different=constrain_different,
        )

        self.combine_by = combine_by
        self.combine_by_powers = combine_by_powers
        self.combine_by_coeffs = combine_by_coeffs
        if self.combine_by == "prod":
            self.combine = self.prod
        elif self.combine_by == "sum":
            self.combine = self.sum
        else:
            raise ValueError(
                'Combine by function not recognized. Must be "sum" or "prod"'
            )

    def prod(self, data_list: List) -> np.ndarray:
        if self.combine_by_powers is not None:
            assert len(self.combine_by_powers) == len(data_list)
            powers = copy.deepcopy(self.combine_by_powers)
        else:
            powers = [1] * len(data_list)

        data_list = copy.deepcopy(data_list)

        # np methods use the __array__ accessor and return W
        value = np.power(data_list.pop(0), powers.pop(0))
        for data, power in zip(data_list, powers):
            value *= np.power(data, power)
        return value

    def sum(self, data_list: List) -> np.ndarray:
        if self.combine_by_powers is not None:
            assert len(self.combine_by_coeffs) == len(data_list)
            coeffs = copy.deepcopy(self.combine_by_coeffs)
        else:
            coeffs = [1] * len(data_list)

        data_list = copy.deepcopy(data_list)

        # np methods use the __array__ accessor and return W
        value = np.multiply(data_list.pop(0), coeffs.pop(0))
        for data, coeff in zip(data_list, coeffs):
            value += np.multiply(data, coeff)

        return value

    def calculate(self, dataset: xr.Dataset) -> Self:
        W_list = []
        for name in listify(self.input_variable):
            W_list.append(dataset[name])

        self.W = self.combine(W_list)
        self.W = self.normalize1()
        self.W = self.apply_constraints()

        dims = [self.sample_dim + "_i", self.sample_dim + "_j"]
        self.output[self.output_variable] = xr.DataArray(self.W, dims=dims)  # type: ignore

        return self



import pandas as pd
import numpy as np
import copy
import scipy.spatial
import random
import logging
from AFL.agent import PhaseMap
import matplotlib.pyplot as plt
from random import shuffle

from sklearn.neighbors import KernelDensity

#move dense_pm definition outside of this class
#move to driver,make settable here and in driver
#pass make to driver

def gauss2d(x=0, y=0, cx=0, cy=0, sx=1, sy=1, a=-1):
    return a / (2. * np.pi * sx * sy) * np.exp(-((x - cx)**2. / (2. * sx**2.) + (y - cy)**2. / (2. * sy**2.)))

class Acquisition:
    def __init__(self):
        self.phasemap = None
        self.mask = None
        self.y_mean = None
        self.y_var = None
        self.next_sample = None
        self.logger = logging.getLogger()
        self.composition_atol = 0.015#absolute distance tolerace
        self.metric_rtol = 0.03 #pct tolerance
    
    def reset_phasemap(self,phasemap):
        self.phasemap = phasemap
        
    def add_exclusion(self,points):
        pass
    
    def plot(self,masked=False,**kwargs):
        if masked:
            pm1 = self.phasemap[[var for var in self.phasemap if ('grid' in self.phasemap[var].dims)]].where(self.phasemap.mask,drop=True).copy()
            pm = self.phasemap.copy()
            pm = pm.drop_dims('grid').update(pm1)
        else:
            pm = self.phasemap.copy()
            
        pm.afl.comp.plot_continuous(components=self.phasemap.attrs['components_grid'],labels='acq_metric')
        pm.afl.comp.plot_discrete(components=self.phasemap.attrs['components'],set_labels=False)
        
        if self.next_sample is not None:
            from AFL.agent.util import ternary_to_xy
            xy = ternary_to_xy(np.array([self.next_sample.squeeze().values]))
            plt.plot(*xy.T,marker='x',color='r')
        return plt.gca()
        
    def copy(self):
        return copy.deepcopy(self)

    def execute(self):
        raise NotImplementedError('Subclasses must implement execute!')

    def get_next_sample(self,nth=0,composition_check=None):

        if np.all(np.isnan(np.unique(self.phasemap['acq_metric']))):
            sample_randomly = True
        else:
            sample_randomly = False

        if 'mask' not in self.phasemap:
            mask = slice(None)
        else:
            mask = self.phasemap.mask
            
        print("Creating ordered metric lists...")
        if sample_randomly:
            close_metric = (
                self.phasemap.sel(grid=mask,drop=False)
                [composition.attrs['components_grid']] #just take composition values
                .rename({k:k.replace('_grid','') for k in composition.attrs['components_grid']})
                .to_array('component')
            )
        else:
            #metric = self.phasemap.sel(grid=mask,drop=True)#.copy()
            metric = self.phasemap.set_index(grid='mask').sel(grid=True,drop=True).reset_index('grid').reset_coords(drop=True)
            metric = metric[['acq_metric']+metric.attrs['components_grid']]
            metric = metric.sortby('acq_metric')
            
            #find samples within rtol of maximum metric value
            is_close = np.isclose(
                metric['acq_metric'],
                metric['acq_metric'].max(),
                rtol=self.metric_rtol,
                atol=0
            )
             
            close_metric = (
                metric.sel(grid=is_close,drop=False)
                [metric.attrs['components_grid']] #just take composition values
                .rename({k:k.replace('_grid','') for k in metric.attrs['components_grid']})
                .to_array('component')
            )

        close_metric_idex_list = list(close_metric.grid.values) #list of integers
        shuffle(close_metric_idex_list)
        
        print(f"Running get_next_sample with sample_randomly={sample_randomly}")
        i = 0
        while True:
            print(f"Running iteration {i}")
            
            if close_metric_idex_list:
                print(f"Getting random point within {self.metric_rtol:0.2f} of maximum...")
                idex = close_metric_idex_list.pop()
            else:
                raise ValueError('No gridpoint found that satisfies constraints! Try increasing composition_rtol!')
            
            composition = close_metric.sel(grid=idex)
            
            if composition_check is None:
                break #all done
        
            print(f"Verifying that gridpoint isn't on top of previous measurement...")
            dist = (composition_check - composition).pipe(np.square).sum('component').pipe(np.sqrt)
            #if (dist<self.composition_atol).any():
            if (dist<self.composition_atol).any():
                i+=1
            else:
                break
        
            if nth>=self.phasemap.grid.shape[0]:
                raise ValueError(f'No next sample found! Searched {nth} iterations!')
            elif nth>1000:
                raise ValueError('Find next sample failed to converge!')

        print(f"Found that gridpoint {idex} satisfies the acquistion function and all constraints")
        self.next_sample = composition
        return self.next_sample

class Variance(Acquisition):
    def __init__(self):
        super().__init__()
        self.name = 'variance'
        
    def calculate_metric(self,GP):
        if self.phasemap is None:
            raise ValueError('No phase map set for acquisition! Call reset_phasemap!')
            
        predict = GP.predict(self.phasemap.attrs['components_grid'])
        self.y_mean = predict['mean']
        self.y_var = predict['var']
        self.phasemap['acq_metric'] = ('grid',GP.y_var.sum(1))
        self.phasemap.attrs['acq_metric'] = self.name

        return self.phasemap
    
class Random(Acquisition):
    def __init__(self):
        super().__init__()
        self.name = 'random'
        
    def calculate_metric(self,GP):
        if self.phasemap is None:
            raise ValueError('No phase map set for acquisition! Call reset_phasemap!')
            
        predict = GP.predict(self.phasemap.attrs['components_grid'])
        self.y_mean = predict['mean']
        self.y_var = predict['var']
            
        indices = np.arange(self.phasemap['grid'].shape[0])
        random.shuffle(indices)
        self.phasemap['acq_metric'] = ('grid',indices)
        self.phasemap.attrs['acq_metric'] = self.name
        return self.phasemap
    
class IterationCombined(Acquisition):
    def __init__(self,function1,function2,function2_frequency=5):
        super().__init__()
        self.function1 = function1
        self.function2 = function2
        # self.name = 'IterationCombined'  
        # self.name += '-'+function1.name
        # self.name += '-'+function2.name
        self.name = function1.name
        self.name += '-'+function2.name
        self.name += '@'+str(function2_frequency)
        self.iteration = 1
        self.function2_frequency=function2_frequency
        
    def reset_phasemap(self,phasemap):
        self.phasemap = phasemap
        self.function1.reset_phasemap(phasemap)
        self.function2.reset_phasemap(phasemap)
        
    def calculate_metric(self,GP):
        if self.function1.phasemap is None:
            raise ValueError('No phase map set for acquisition! Call reset_phasemap!')

        predict = GP.predict(self.phasemap.attrs['components_grid'])
        self.y_mean = predict['mean']
        self.y_var = predict['var']
        
        if ((self.iteration%self.function2_frequency)==0):
            print(f'Using acquisition function {self.function2.name} of iteration {self.iteration}')
            self.phasemap = self.function2.calculate_metric(GP)
            self.phasemap.attrs['acq_current_metric'] = self.function2.name
        else:
            print(f'Using acquisition function {self.function1.name} of iteration {self.iteration}')
            self.phasemap = self.function1.calculate_metric(GP)
            self.phasemap.attrs['acq_current_metric'] = self.function1.name

        self.phasemap.attrs['acq_metric'] = self.name
        self.phasemap.attrs['acq_metric1'] = self.function1.name
        self.phasemap.attrs['acq_metric2'] = self.function2.name
        self.phasemap.attrs['acq_iteration'] = self.iteration
        self.iteration+=1
            
        return self.phasemap
    
class LowestDensity(Acquisition):
    def __init__(self,bandwidth=0.075):
        super().__init__()
        self.name = 'LowestDensity'
        self.bandwidth=bandwidth
        
    def calculate_metric(self,GP=None,y_mean=None,y_var=None): 
        if self.phasemap is None:
            raise ValueError('No phase map set for acquisition! Call reset_phasemap!')
            
        #this must be calculated regardless of whether it is used
        if GP is not None:
            predict = GP.predict(self.phasemap.attrs['components_grid'])
            self.y_mean = predict['mean']
            self.y_var = predict['var']
        elif (y_mean is not None) and (y_var is not None):
            self.y_mean = y_mean
            self.y_var = y_var
        else:
            raise ValueError('Need to pass either GP or y_mean/y_var into calcualte_metric!')
            

        GP_domain_transform = self.phasemap.attrs.get('GP_domain_transform',None)
        if GP_domain_transform=='ternary':
            xy = self.phasemap.afl.comp.to_ternary_xy()
            xy_grid = self.phasemap.afl.comp.to_xy(self.phasemap.attrs['components_grid'])
        # elif GP_domain_transform=='range_scaled':
        #     ranges = {}
        #     for component in components:
        #         ranges[component] = self.dataset.attrs[component+'_range'][1] - self.dataset.attrs[component+'_range'][0]
        #     xy  = self.dataset.afl.comp.get_range_scaled(ranges=ranges,components=self.phasemap.attrs['components'])
        #     xy_grid = self.dataset.afl.comp.get_range_scaled(ranges=ranges,components=self.phasemap.attrs['components_grid'])
        else:
            xy = self.phasemap[self.phasemap.attrs['components']].to_array('component').transpose(...,'component')
            xy_grid = self.phasemap[self.phasemap.attrs['components_grid']].to_array('component').transpose(...,'component')

        kde = KernelDensity(bandwidth=self.bandwidth)
        kde.fit(xy)
        
        acq_metric = -np.exp(kde.score_samples(xy_grid))
        self.phasemap['acq_metric'] = ('grid',acq_metric)
        self.phasemap.attrs['acq_metric'] = 'LowestDensity'
        return self.phasemap



import pandas as pd
import numpy as np
import copy
import scipy.spatial
import random
import logging
from AFL.agent import PhaseMap
import matplotlib.pyplot as plt
from random import shuffle

from sklearn.neighbors import KernelDensity

#move dense_pm definition outside of this class
#move to driver,make settable here and in driver
#pass make to driver

def gauss2d(x=0, y=0, cx=0, cy=0, sx=1, sy=1, a=-1):
    return a / (2. * np.pi * sx * sy) * np.exp(-((x - cx)**2. / (2. * sx**2.) + (y - cy)**2. / (2. * sy**2.)))

class Acquisition:
    def __init__(self):
        self.phasemap = None
        self.mask = None
        self.y_mean_GPR = None
        self.y_var_GPR = None
        self.y_mean_GPC = None
        self.y_var_GPC = None
        self.next_sample = None
        self.logger = logging.getLogger()
        self.composition_atol = 0.015#absolute distance tolerace
        self.metric_rtol = 0.03 #pct tolerance
    
    def reset_phasemap(self,phasemap):
        self.phasemap = phasemap
        
    def add_exclusion(self,points):
        pass
    
    def plot(self,masked=False,**kwargs):
        if masked:
            pm1 = self.phasemap[[var for var in self.phasemap if ('grid' in self.phasemap[var].dims)]].where(self.phasemap.mask,drop=True).copy()
            pm = self.phasemap.copy()
            pm = pm.drop_dims('grid').update(pm1)
        else:
            pm = self.phasemap.copy()
            
        pm.afl.comp.plot_continuous(components=self.phasemap.attrs['components_grid'],labels='acq_metric')
        pm.afl.comp.plot_discrete(components=self.phasemap.attrs['components'],set_labels=False)
        
        if self.next_sample is not None:
            plt.plot(*PhaseMap.to_xy(np.array([self.next_sample.squeeze().values])).T,marker='x',color='r')
        return plt.gca()
        
    def copy(self):
        return copy.deepcopy(self)

    def execute(self):
        raise NotImplementedError('Subclasses must implement execute!')

    def get_next_sample(self,nth=0,composition_check=None, sample_randomly=False):

        # if np.all(np.isnan(np.unique(self.phasemap['acq_metric']))):
        #     sample_randomly = True
        # else:
        #     sample_randomly = False
        
        if 'mask' not in self.phasemap:
            mask = slice(None)
        else:
            mask = self.phasemap.mask
            
        print("Creating ordered metric lists...")
        print("are you sampling randomly? ", sample_randomly)
        if sample_randomly:
            print("sampling randomly")
            close_metric = (
                self.phasemap.sel(grid=mask,drop=False)
                [self.phasemap.attrs['components_grid']] #just take composition values
                .rename({k:k.replace('_grid','') for k in self.phasemap.attrs['components_grid']})
                .to_array('component')
            )
            
            
        else:
            print('not sampling randomly')
            # metric = self.phasemap.sel(grid=mask,drop=True)#.copy()
            metric = self.phasemap.set_index(grid='mask').sel(grid=True,drop=True).reset_index('grid').reset_coords(drop=True)
            metric = metric[['acq_metric']+metric.attrs['components_grid']]
            metric = metric.sortby('acq_metric')
            
            #find samples within rtol of maximum metric value
            is_close = np.isclose(
                metric['acq_metric'],
                metric['acq_metric'].max(),
                rtol=self.metric_rtol,
                atol=0
            )
             
            close_metric = (
                metric.sel(grid=is_close,drop=False)
                [metric.attrs['components_grid']] #just take composition values
                .rename({k:k.replace('_grid','') for k in metric.attrs['components_grid']})
                .to_array('component')
            )

        close_metric_idex_list = list(close_metric.grid.values) #list of integers
        shuffle(close_metric_idex_list)
        
        print(f"Running get_next_sample with sample_randomly={sample_randomly}")
        i = 0
        while True:
            print(f"Running iteration {i}")
            
            if close_metric_idex_list:
                print(f"Getting random point within {self.metric_rtol:0.2f} of maximum...")
                idex = close_metric_idex_list.pop()
            else:
                raise ValueError('No gridpoint found that satisfies constraints! Try increasing composition_rtol!')
                
            
            composition = close_metric.sel(grid=idex)
            
            if composition_check is None:
                break #all done
        
            print(f"Verifying that gridpoint isn't on top of previous measurement...")
            dist = (composition_check - composition).pipe(np.square).sum('component').pipe(np.sqrt)
            #if (dist<self.composition_atol).any():
            if (dist<self.composition_atol).any():
                i+=1
            else:
                break
        
            if nth>=self.phasemap.grid.shape[0]:
                raise ValueError(f'No next sample found! Searched {nth} iterations!')
            elif nth>1000:
                raise ValueError('Find next sample failed to converge!')

        print(f"Found that gridpoint {idex} satisfies the acquistion function and all constraints")
        self.next_sample = composition
        return self.next_sample
    
def get_next_sample_passthrough(self,next_sample,*args,**kwargs):
    return next_sample

class pseudoUCB(Acquisition):
    """
    This acquisition function works on two GPs only. One has to be a classifier and the other has to be a regressor
    This needs to be generalizable and requires some thought...
    """
    def __init__(self,scaling =1.0, Thompson_sampling=False):
        super().__init__()
        self.name = 'pseudo UCB'
        self.scaling = scaling
        self.Thompson_sampling = Thompson_sampling
        self.regressor_sample = None

    def calculate_metric(self,GP,GPR):
        if self.phasemap is None:
            raise ValueError('No phase maps set for acquisition! Call reset_phasemap!')

        #needs to have some flexibility. for every GP or GPR in this calculation, 
        classifier_prediction = GP.predict(self.phasemap.attrs['components_grid'])
        self.y_mean_GPC = classifier_prediction['mean']
        self.y_var_GPC = classifier_prediction['var']

        y_var_GPC_scaled =  self.y_var_GPC.sum(1)
        print('min max of model variance')
        print(y_var_GPC_scaled.min(), y_var_GPC_scaled.max())
#         y_var_GPC_scaled = (y_var_GPC_scaled - y_var_GPC_scaled.mean())/y_var_GPC_scaled.std()

        
        regressor_prediction = GPR.predict(self.phasemap.attrs['components_grid'])
        self.y_mean_GPR = regressor_prediction['mean']
        self.y_var_GPR  = regressor_prediction['var']

        y_mean_GPR_scaled =  self.y_mean_GPR.squeeze()
        print('min max of model mean')
        print(y_mean_GPR_scaled.min(), y_mean_GPR_scaled.max())
#         y_mean_GPR_scaled = (y_mean_GPR_scaled - y_mean_GPR_scaled.mean())/y_mean_GPR_scaled.std()
        
        if self.Thompson_sampling:
            self.regressor_sample = GPR.model.predict_f_samples(GPR.transform_domain(components=self.phasemap.attrs['components_grid']),1)
            self.regressor_sample = self.regressor_sample.numpy().squeeze()
            regressor_sample_scaled = self.regressor_sample
            # regressor_sample_scaled = (self.regressor_sample - y_mean_GPR_scaled.mean())/y_mean_GPR_scaled.std()
            self.phasemap['acq_metric'] = ('grid', regressor_sample_scaled + self.scaling * y_var_GPC_scaled)
            self.phasemap.attrs['acq_metric'] = self.name + '_TS'
        else:
            self.phasemap['acq_metric'] = ('grid',y_mean_GPR_scaled + self.scaling * y_var_GPC_scaled)
            self.phasemap.attrs['acq_metric'] = self.name
        self.phasemap.attrs['UCB_scaling'] = self.scaling

        return self.phasemap

# pseudoUCB.get_next_sample=get_next_sample_noThompson


# def get_next_sample_noThompson


"""
Pipeline module for the double_agent package.

This module provides the Pipeline class, which serves as a container for defining
and executing computational workflows. Pipelines are composed of PipelineOp objects
that perform specific operations on data.

PipelineOps in this system follow these conventions:
- Each operator takes input_variable and output_variable parameters in its constructor
- Each operator implements a .calculate method that:
  - Maintains the same signature as the base class (PipelineOp)
  - Writes results to an xarray Dataset or DataArray
  - Returns self for method chaining
"""

import copy
import pickle
import re
from typing import Generator, Optional, List
import warnings
import datetime
import json
import pathlib

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import xarray as xr
from tqdm.auto import tqdm
from typing_extensions import Self

from AFL.double_agent.PipelineContext import PipelineContext
from AFL.double_agent.PipelineOp import PipelineOp
from AFL.double_agent.util import listify


class Pipeline(PipelineContext):
    """
    Container class for defining and executing computational workflows.

    The Pipeline class serves as a framework for organizing and running sequences of
    operations (PipelineOps) on data. Each operation in the pipeline takes input data,
    performs a specific transformation, and produces output data that can be used by
    subsequent operations.

    Parameters
    ----------
    name : Optional[str], default=None
        Name of the pipeline. If None, defaults to "Pipeline".
    ops : Optional[List], default=None
        List of PipelineOp objects to initialize the pipeline with.

    Attributes
    ----------
    result : Any
        Stores the final result after pipeline execution
    ops : List
        List of PipelineOp objects in the pipeline
    graph : nx.DiGraph
        NetworkX directed graph representation of the pipeline
    graph_edge_labels : Dict
        Edge labels for the pipeline graph visualization
    """

    def __init__(
        self,
        name: Optional[str] = None,
        ops: Optional[List] = None,
        description: Optional[str] = None,
    ) -> None:
        self.result = None
        self.ops = ops or []
        self.description = str(description)
        self.name = name or "Pipeline"

        # placeholder for networkx graph
        self.graph = None
        self.graph_edge_labels = None

    def __iter__(self) -> Generator[PipelineOp, Self, None]:
        """Allows pipelines to be iterated over"""
        for op in self.ops:
            yield op

    def __getitem__(self, i: int) -> PipelineOp:
        """Allows PipelineOps to be viewed from the pipeline via an index"""
        return self.ops[i]

    def __repr__(self) -> str:
        return f"<Pipeline {self.name} N={len(self.ops)}>"

    def search(self, name: str, contains: bool = False) -> Optional[PipelineOp]:
        for op in self:
            if contains and (name in op.name):
                return op
            elif name == op.name:
                return op
        return None

    def print(self) -> None:
        """Print a summary of the pipeline"""
        print(f"{'PipelineOp':40s} {'input_variable'} ---> {'output_variable'}")
        print(f"{'-' * 10:40s} {'-' * 35}")
        for i, op in enumerate(self):
            print(
                f"{i:<3d}) {'<' + op.name + '>':35s} {op.input_variable} ---> {op.output_variable}"
            )

        print()
        print("Input Variables")
        print("---------------")
        for i, data in enumerate(self.input_variables()):
            print(f"{i}) {data}")

        print()
        print("Output Variables")
        print("----------------")
        for i, data in enumerate(self.output_variables()):
            print(f"{i}) {data}")

    def print_code(self) -> None:
        """String representation of approximate code to generate this pipeline

        Run this method to produce a string of Python code that should
        recreate this Pipeline.

        """

        output_string = f'with Pipeline(name = "{self.name}") as p:\n'

        for op in self:
            args = op._stored_args
            output_string += f"    {type(op).__name__}(\n"
            for k, v in args.items():
                if isinstance(v, str):
                    output_string += f'        {k}="{v}",\n'
                else:
                    output_string += f"        {k}={v},\n"
            output_string += f"    )\n\n"

        try:
            shell = get_ipython().__class__.__name__
            if shell == "ZMQInteractiveShell":
                # Create IPython magic for creating executable code
                ip = get_ipython()

                # First display the code with syntax highlighting for visibility
                from IPython.display import display, Code

                # display(Code(output_string, language="python"))

                # Define a temporary magic to create a new cell with the code
                ip.set_next_input(output_string)
                print("Pipeline code has been prepared in a new cell below.")
            else:
                print(output_string)
        except NameError:
            print(output_string)

    def append(self, op: PipelineOp) -> Self:
        """Mirrors the behavior of python lists"""
        self.ops.append(op)
        return self

    def extend(self, op: List[PipelineOp]) -> Self:
        """Mirrors the behavior of python lists"""
        self.ops.extend(op)
        return self

    def clear_outputs(self):
        """Clears the output dict of all PipelineOps in this pipeline"""
        for op in self:
            op.output = {}

    def copy(self) -> Self:
        return copy.deepcopy(self)

    def write_json(
        self, filename: str, overwrite=False, description: Optional[str] = None
    ):
        """Write pipeline to disk as a JSON

        Parameters
        ----------
        filename: str
            Filename or filepath to be written
        overwrite: bool, default=False
            Whether to overwrite an existing file
        description: str, optional
            A descriptive text about the pipeline's purpose and functionality
        """

        if not overwrite and pathlib.Path(filename).exists():
            raise FileExistsError()

        pipeline_dict = {
            "name": self.name,
            "date": datetime.datetime.now().strftime("%m/%d/%y %H:%M:%S-%f"),
            "description": (
                str(description) if description is not None else self.description
            ),
            "ops": [op.to_json() for op in self],
        }

        with open(filename, "w") as f:
            json.dump(pipeline_dict, f, indent=1)

        print(f"Pipeline successfully written to {filename}.")

    @staticmethod
    def read_json(filename: str):
        """Read pipeline from json file on disk

        Usage
        -----
        ```python
        from AFL.double_agent.Pipeline import Pipeline
        pipeline1 = Pipeline.read_json('pickled_pipeline.pkl')
        ````
        """
        with open(filename, "r") as f:
            pipeline_dict = json.load(f)

        pipeline = Pipeline(
            name=pipeline_dict["name"],
            ops=[PipelineOp.from_json(op) for op in pipeline_dict["ops"]],
            description=pipeline_dict["description"],
        )

        return pipeline

    def write_pkl(self, filename: str):
        """Write pipeline to disk as a pkl

        .. warning::
            Please use the read_json and write_json methods. The pickle methods
            are insecure and prone to errors.

        Parameters
        ----------
        filename: str
            Filename or filepath to be written
        """
        pipeline = self.copy()
        pipeline.clear_outputs()

        with open(filename, "wb") as f:
            pickle.dump(pipeline, f)

    @staticmethod
    def read_pkl(filename: str):
        """Read pipeline from pickle file on disk

        .. warning::
            Please use the  `read_json`  and `write_json` methods. The pickle methods
            are insecure and prone to errors.

        Usage
        -----
        ```python
        from AFL.double_agent.Pipeline import Pipeline
        pipeline1 = Pipeline.read('pickled_pipeline.pkl')
        ````

        """
        with open(filename, "rb") as f:
            pipeline = pickle.load(f)
        return pipeline

    def make_graph(self):
        """Build a networkx graph representation of this pipeline"""

        self.graph = nx.DiGraph()
        self.graph_edge_labels = {}
        for op in self:
            for output_variable in listify(op.output_variable):
                if output_variable is None:
                    continue
                self.graph.add_node(output_variable)
                # need to handle case where input_variables is a list
                for input_variable in listify(op.input_variable):
                    if input_variable is None:
                        continue
                    self.graph.add_node(input_variable)
                    self.graph.add_edge(input_variable, output_variable, name=op.name)
                    self.graph_edge_labels[input_variable, output_variable] = op.name

    def draw(self, figsize=(8, 8), edge_labels=True):
        """Draw the pipeline as a graph"""
        self.make_graph()

        fig = plt.figure(figsize=figsize)
        pos = nx.nx_agraph.pygraphviz_layout(self.graph, prog="dot")
        nx.draw(self.graph, with_labels=True, pos=pos, node_size=1000)
        if edge_labels:
            nx.draw_networkx_edge_labels(self.graph, pos, self.graph_edge_labels)
        return fig

    def draw_plotly(self):
        import plotly.graph_objects as go

        self.make_graph()
        pos = nx.nx_agraph.pygraphviz_layout(self.graph, prog="dot")

        edge_x = []
        edge_y = []
        for edge in self.graph.edges():
            x0, y0 = pos[edge[0]]
            x1, y1 = pos[edge[1]]
            edge_x.append(x0)
            edge_x.append(x1)
            edge_x.append(None)
            edge_y.append(y0)
            edge_y.append(y1)
            edge_y.append(None)

        edge_trace = go.Scatter(
            x=edge_x,
            y=edge_y,
            line=dict(width=0.5, color="#888"),
            hoverinfo="none",
            mode="lines",
        )

        node_x = []
        node_y = []
        for node in self.graph.nodes():
            x, y = pos[node]
            node_x.append(x)
            node_y.append(y)

        node_trace = go.Scatter(
            x=node_x,
            y=node_y,
            mode="markers",
            hoverinfo="text",
            marker=dict(reversescale=True, color=[], size=20, line_width=2),
        )

        node_text = []
        regex = re.compile("[\[\]']")
        for node in self.graph.nodes():
            VarIn = np.unique(list(n[0] for n in self.graph.in_edges(node)))
            VarOut = np.unique(list(n[0] for n in self.graph.out_edges(node)))
            OpIn = np.unique([self.graph_edge_labels[i, node] for i in VarIn])

            OpIn = regex.sub("", str(OpIn)).replace(" ", ", ")
            VarIn = regex.sub("", str(VarIn)).replace(" ", ", ")
            VarOut = regex.sub("", str(VarOut)).replace(" ", ", ")
            node_text.append(
                f"{'Node: ':9s}{node}<br>{'OpIn: ':9s}{OpIn}<br>{'VarIn: ':9s}{VarIn}<br>{'VarOut: ':9s}{VarOut}"
            )

        node_trace.text = node_text

        fig = go.Figure([node_trace, edge_trace])
        fig.update_layout(
            height=750,
            width=750,
            showlegend=False,
            template="simple_white",
            xaxis={"visible": False},
            yaxis={"visible": False},
            margin=dict(l=20, r=20, t=20, b=20),
        )
        return fig

    def input_variables(self) -> List[str]:
        """Get the input variables needed for the pipeline to fully execute

        Warning
        -------
        This list will currently include "Generators" that aren't required to be in the pipeline. These will hopefully
        be distinguishable by having Generator in the name, but this isn't enforced at the moment.
        """
        self.make_graph()
        return [n for n, d in self.graph.in_degree() if d == 0]  # type: ignore

    def output_variables(self) -> List[str]:
        """Get the outputs variables of the pipeline"""
        self.make_graph()
        return [n for n, d in self.graph.out_degree() if d == 0]  # type: ignore

    def calculate(
        self, dataset: xr.Dataset, tiled_data=None, disable_progress_bar: bool = False
    ) -> xr.Dataset:
        """Execute all operations in pipeline on provided dataset"""
        dataset1 = dataset.copy()

        for op in (pbar := tqdm(self.ops, disable=disable_progress_bar)):
            pbar.set_postfix_str(f"{op.name:25s}")
            op.calculate(dataset1)
            dataset1 = op.add_to_dataset(dataset1, copy_dataset=False)

            if tiled_data is not None:
                op.add_to_tiled(tiled_data)

        self.result = dataset1
        return dataset1



import itertools

import xarray as xr
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter
from sklearn.preprocessing import OrdinalEncoder
import warnings

import textwrap

from AFL.agent.util import ternary_to_xy, composition_grid, composition_grid_ternary, mpl_format_ternary

from sklearn.preprocessing import StandardScaler


@xr.register_dataarray_accessor('afl')
class AFL_DataArrayTools:
    def __init__(self, da):
        self.da = da
        self.comp = CompositionTools(da)
        self.scatt = ScatteringTools(da)
        self.util = Utilities(da)


@xr.register_dataset_accessor('afl')
class AFL_DatasetTools:
    def __init__(self, ds):
        self.ds = ds
        self.comp = CompositionTools(ds)
        self.labels = LabelTools(ds)

    def append(self, data_dict, concat_dim='sample'):
        '''Append data to current dataset (warning: much copying and data loss)'''

        # need to make sure all DataArrays have concat_dim
        # initial pass to prep for new Dataset creation
        for k in data_dict.keys():
            # make sure all variables in data_dict are already in this dataset
            if k not in self.ds:
                raise ValueError(textwrap.fill(textwrap.dedent(f'''
                All variables in data_dict {list(data_dict.keys())} must be
                in this Dataset before appending
                ''')))

            # make sure all DataArrays have concat_dim
            v = data_dict[k]
            if isinstance(v, xr.DataArray):
                if concat_dim not in v.coords:
                    data_dict[k] = xr.DataArray([v], dims=[concat_dim, *v.dims], coords=v.coords)

        ds_new = xr.Dataset()
        for k, v in data_dict.items():
            if isinstance(v, xr.DataArray):
                ds_new[k] = xr.concat([self.ds[k].copy(), v.copy()], dim=concat_dim)
            else:
                v_xr = xr.DataArray([v], dims=concat_dim)
                ds_new[k] = xr.concat([self.ds[k].copy(), v_xr], dim=concat_dim)
        return ds_new


class LabelTools:
    def __init__(self, data):
        self.data = data

    def make_default(self, name='labels', dim='sample'):
        data = self.data.copy()
        data['labels'] = xr.DataArray(np.ones(self.data[dim].shape[0]), dims=dim, coords=self.data[dim].coords)
        return data

    def make_ordinal(self, ordinal_name='labels_ordinal', labels_name='labels', sortby=None):
        encoder = OrdinalEncoder()
        data = self.data.copy()
        if sortby is not None:
            data = data.sortby(sortby)
        labels_ordinal = encoder.fit_transform(data[labels_name].values.reshape(-1, 1)).flatten()
        data[ordinal_name] = data[labels_name].copy(data=labels_ordinal)
        return data


class ScatteringTools:
    def __init__(self, data):
        self.data = data

    def plot_logwaterfall(self, x='q', dim='sample', ylabel='Intensity [A.U.]', xlabel=r'q [$Å^{-1}$]', legend=True,
                          base=10, ax=None):
        N = self.data[dim].shape[0]
        mul = self.data[dim].copy(data=np.geomspace(1, float(base) ** (N + 1), N))
        lines = (
            self
            .data
            .pipe(lambda x: x * mul)
            .plot
            .line(
                x=x,
                xscale='log',
                yscale='log',
                marker='.',
                ls='None',
                add_legend=legend,
                ax=ax)
        )
        plt.gca().set(xlabel=xlabel, ylabel=ylabel, )
        if legend and (len(lines) > 1):
            sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
        return lines

    def plot_waterfall(self, x='logq', dim='sample', ylabel='Intensity [A.U.]', xlabel=r'q [$Å^{-1}$]', legend=True,
                       base=1, ax=None):
        N = self.data[dim].shape[0]
        mul = self.data[dim].copy(data=np.linspace(1, base * N, N))
        lines = (self
        .data
        .pipe(lambda x: x + mul)
        .plot
        .line(
            x=x,
            xscale='linear',
            yscale='linear',
            marker='.',
            ls='None',
            add_legend=legend,
            ax=ax)
        )
        plt.gca().set(xlabel=xlabel, ylabel=ylabel, )
        if legend and (len(lines) > 1):
            sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
        return lines

    def plot_loglog(self, x='q', ylabel='Intensity [A.U.]', xlabel=r'q [$Å^{-1}$]', legend=True, ax=None, **mpl_kw):
        lines = self.data.plot.line(
            x=x,
            xscale='log',
            yscale='log',
            marker='.',
            ls='None',
            add_legend=legend,
            ax=ax,
            **mpl_kw)
        plt.gca().set(xlabel=xlabel, ylabel=ylabel)
        if legend and (len(lines) > 1):
            sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
        return lines

    def plot_linlin(self, x='logq', ylabel='Intensity [A.U.]', xlabel=r'q [$Å^{-1}$]', legend=True, ax=None, **mpl_kw):
        lines = self.data.plot.line(
            x=x,
            marker='.',
            ls='None',
            add_legend=legend,
            ax=ax,
            **mpl_kw)
        plt.gca().set(xlabel=xlabel, ylabel=ylabel)
        if legend and (len(lines) > 1):
            sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
        return lines

    def clean(self, qlo=None, qhi=None, qlo_isel=None, qhi_isel=None, pedestal=1e-12, npts=250, derivative=0,
              sgf_window_length=31, sgf_polyorder=2, apply_log_scale=True):
        data2 = self.data.copy()
        data2.name = 'I'

        if ((qlo is not None) or (qhi is not None)) and ((qlo_isel is not None) or (qhi_isel is not None)):
            warnings.warn((
                'Both value and index based indexing have been specified! '
                'Value indexing will be applied first!'
            )
                , stacklevel=2)

        data2 = data2.sel(q=slice(qlo, qhi))
        data2 = data2.isel(q=slice(qlo_isel, qhi_isel))

        if pedestal is not None:
            data2 += pedestal

        qname = 'logq'
        if apply_log_scale:
            data2 = data2.where(data2 > 0.0, drop=True)
            data2 = data2.pipe(np.log10)
            data2['q'] = np.log10(data2.q)
            data2 = data2.rename(q=qname)

        # need to remove duplicate values
        data2 = data2.groupby(qname, squeeze=False).mean()

        # set minimum value of scattering to pedestal value and fill nans with this value
        if pedestal is not None:
            data2 += pedestal
            data2 = data2.where(~np.isnan(data2)).fillna(pedestal)

        # interpolate to constant log(dq) grid
        qnew = np.geomspace(data2[qname].min(), data2[qname].max(), npts)
        dq = qnew[1] - qnew[0]
        data2 = data2.interp({qname: qnew})

        # filter out any q that have NaN
        data2 = data2.dropna(qname, 'any')

        # take derivative
        dI = savgol_filter(data2.values.T, window_length=sgf_window_length, polyorder=sgf_polyorder, delta=dq, axis=0,
                           deriv=derivative)
        data2_dI = data2.copy(data=dI.T)
        return data2_dI


class Utilities:
    def __init__(self, data):
        self.data = data

    def preprocess_data(self,
                        xname,
                        yname,
                        xlo=None,
                        xhi=None,
                        xlo_isel=None,
                        xhi_isel=None,
                        pedestal=None,
                        npts=250,
                        derivative=0,
                        sgf_window_length=31,
                        sgf_polyorder=2,
                        logx=False,
                        logy=False
                        ):
        data2 = self.data.copy()
        data2.name = yname

        if ((xlo is not None) or (xhi is not None)) and ((xlo_isel is not None) or (xhi_isel is not None)):
            warnings.warn((
                'Both value and index based indexing have been specified! '
                'Value indexing will be applied first!'
            )
                , stacklevel=2)

        data2 = data2.sel(**{xname: slice(xlo, xhi)})
        data2 = data2.isel(**{xname: slice(xlo_isel, xhi_isel)})

        if logy:
            log_xname = 'log' + xname
            data2 = data2.where(data2 > 0.0, drop=True)
            data2 = data2.pipe(np.log10)
            data2[xname] = np.log10(data2[xname])
            data2 = data2.rename(**{xname: log_xname})

            data2 = data2.groupby(log_xname, squeeze=False).mean()
        else:
            # need to remove duplicate values
            data2 = data2.groupby(xname, squeeze=False).mean()

        # set minimum value of scattering to pedestal value and fill nans with this value
        if pedestal is not None:
            data2 += pedestal
            data2 = data2.where(~np.isnan(data2)).fillna(pedestal)

        # interpolate to constant log(dq) grid
        if logx:
            xnew = np.geomspace(data2[log_xname].min(), data2[log_xname].max(), npts)
            dx = xnew[1] - xnew[0]
            data2 = data2.interp({log_xname: xnew})
            data2 = data2.dropna(log_xname, 'any')
        else:
            xnew = np.linspace(data2[xname].min(), data2[xname].max(), npts)
            dx = xnew[1] - xnew[0]
            data2 = data2.interp({xname: xnew})
            data2 = data2.dropna(xname, 'any')

        # take derivative
        dy = savgol_filter(data2.values.T, window_length=sgf_window_length, polyorder=sgf_polyorder, delta=dx, axis=0,
                           deriv=derivative)
        data2_dy = data2.copy(data=dy.T)
        return data2_dy


class CompositionTools:
    def __init__(self, data):
        self.data = data
        self.cmap = 'viridis'
        self.ranges = {}

    def _get_default(self, components=None, add_grid=False):
        if (components is None):
            try:
                components = self.data.attrs['components']
            except KeyError:
                raise ValueError('Must pass components or set "components" in Dataset attributes.')
        if add_grid:
            components = [c if 'grid' in c else c + '_grid' for c in components]
        return components

    def get_standard_scaled(self, components=None):
        comp = self.get(components)
        comp_scaled = comp.copy(data=StandardScaler().fit_transform(comp))
        return comp_scaled

    def get_range_scaled(self, spec, components=None):
        components = self._get_default(components)
        comp = self.get(components)

        for component in components:
            if component not in spec:
                raise ValueError(
                    'Cannot scale components without grid ranges specified. Call dataset.afl.comp.set_grid_range()')
            comp.loc[dict(component=component)] = comp.sel(component=component).pipe(
                lambda x: (x - spec[component]['min']) / (spec[component]['range']))
        return comp

    def get(self, components=None):
        components = self._get_default(components)

        comp = self.data[components].to_array('component')
        comp.name = 'compositions'
        return comp.transpose(..., 'component')

    def get_grid_old(self, components=None):
        components = self._get_default(components, add_grid=True)

        da_list = []
        for k in components:
            da_list.append(self.data[k + '_grid'])
        comp = xr.concat(da_list, dim='component')
        comp.name = 'compositions_grid'
        comp = comp.assign_coords(component=components)
        return comp.transpose()  # ensures columns are components

    def get_grid(self, components=None):
        components = self._get_default(components, add_grid=False)
        components_grid = self._get_default(components, add_grid=True)

        comp = self.data.set_index(grid=components_grid).grid  # .transpose(...,'component')
        values = np.array([np.array(i) for i in comp.values])  # need to convert to numpy array
        comp = comp.expand_dims(axis=1, component=len(components)).copy(data=values)
        comp = comp.assign_coords(component=components).transpose(..., 'component')
        comp.name = 'composition_grid'
        return comp

    def add_grid(self, components=None, pts_per_row=50, basis=1.0, dim_name='grid', ternary=False, overwrite=False):
        components = self._get_default(components)
        print(f'--> Making grid for components {components} at {pts_per_row} pts_per_row')

        if ternary:
            compositions = composition_grid_ternary(
                pts_per_row=pts_per_row,
                basis=basis,
                dim=len(components),
            )
        else:
            range_spec = {}
            for component in components:
                range_spec[component] = {
                    # 'min':self.data.attrs[component+'_grid_range'][0],
                    # 'max':self.data.attrs[component+'_grid_range'][1],
                    'min': self.data.attrs[component + '_range'][0],
                    'max': self.data.attrs[component + '_range'][1],
                    'steps': self.data.attrs.get(component + '_npts', pts_per_row)
                }
            compositions = composition_grid(components, range_spec)

        for component in components:
            name = component + '_grid'
            if name in self.data:
                if overwrite:
                    del self.data[name]
                else:
                    raise ValueError(f'Component {name} already in Dataset and overwrite is False.')

        components_grid = []
        for component, comps in zip(components, compositions.T):
            name = component + '_grid'
            self.data[name] = (dim_name, comps)
            components_grid.append(name)
        self.data.attrs['components_grid'] = components_grid
        if not ternary:
            for component in components:
                self.data.attrs[component + '_grid_range'] = self.data.attrs[component + '_range']
        return self.data

    def plot_surface(self, components=None, labels=None, set_axes_labels=True, ternary=True, **mpl_kw):
        components = self._get_default(components)

        if len(components) == 3 and ternary:
            try:
                import mpltern
            except ImportError as e:
                raise ImportError('Could not import mpltern. Please install via conda or pip') from e
            projection = 'ternary'
        elif len(components) == 2:
            projection = None
        else:
            raise ValueError(f'plot_surface only compatible with 2 or 3 components. You passed: {components}')

        # coords = np.vstack(list(self.data[c].values for c in components)).T
        coords = self.data[components].to_array('component').transpose(..., 'component').values

        if (labels is None):
            if ('labels' in self.data.coords):
                labels = self.data.coords['labels'].values
            elif ('labels' in self.data):
                labels = self.data['labels'].values
            else:
                labels = np.zeros(coords.shape[0])
        elif isinstance(labels, str) and (labels in self.data):
            labels = self.data[labels].values

        if ('cmap' not in mpl_kw) and ('color' not in mpl_kw):
            mpl_kw['cmap'] = self.cmap
        if 'edgecolor' not in mpl_kw:
            mpl_kw['edgecolor'] = 'face'

        if ('ax' in mpl_kw):
            ax = mpl_kw.pop('ax')

        else:
            fig, ax = plt.subplots(1, 1, subplot_kw=dict(projection=projection))

        artists = ax.tripcolor(*coords.T, labels, **mpl_kw)

        if set_axes_labels:
            if projection == 'ternary':
                labels = {k: v for k, v in zip(['tlabel', 'llabel', 'rlabel'], components)}
            else:
                labels = {k: v for k, v in zip(['xlabel', 'ylabel'], components)}
            ax.set(**labels)
            ax.grid('on', color='black')
        return artists

    def plot_continuous(self, components=None, labels=None, set_labels=True, ternary=True, **mpl_kw):
        warnings.warn('plot_continuous is deprecated and will be removed in a future release. Please use plot_surface.',
                      DeprecationWarning, stacklevel=2)

        components = self._get_default(components)

        if ternary and len(components) == 3:
            xy = self.ternary_to_xy(components)
        elif len(components) == 3:
            xy = np.vstack(list(self.data[c].values for c in components)).T
        elif len(components) == 2:
            xy = np.vstack(list(self.data[c].values for c in components)).T
        else:
            raise ValueError(f'Can only work with 2 or 3 components. You passed: {components}')

        if (labels is None):
            if ('labels' in self.data.coords):
                labels = self.data.coords['labels'].values
            elif ('labels' in self.data):
                labels = self.data['labels'].values
            else:
                labels = np.zeros(xy.shape[0])
        elif isinstance(labels, str) and (labels in self.data):
            labels = self.data[labels].values

        if ('cmap' not in mpl_kw) and ('color' not in mpl_kw):
            mpl_kw['cmap'] = 'viridis'
        if 'color' not in mpl_kw:
            mpl_kw['c'] = labels
        artists = plt.scatter(*xy.T, **mpl_kw)

        if set_labels:
            if ternary:
                mpl_format_ternary(plt.gca(), *components)
            else:
                plt.gca().set(xlabel=components[0], ylabel=components[1])
        return artists

    def plot_scatter(self, components=None, labels=None, set_axes_labels=True, ternary=True, discrete_labels=True,
                     **mpl_kw):
        components = self._get_default(components)

        if len(components) == 3 and ternary:
            try:
                import mpltern
            except ImportError as e:
                raise ImportError('Could not import mpltern. Please install via conda or pip') from e

            projection = 'ternary'
        elif len(components) == 3:
            projection = None
        elif len(components) == 2:
            projection = None
        else:
            raise ValueError(f'plot_surface only compatible with 2 or 3 components. You passed: {components}')

        coords = np.vstack(list(self.data[c].values for c in components)).T

        if (labels is None):
            if ('labels' in self.data.coords):
                labels = self.data.coords['labels'].values
            elif ('labels' in self.data):
                labels = self.data['labels'].values
            else:
                labels = np.zeros(coords.shape[0])
        elif isinstance(labels, str) and (labels in self.data):
            labels = self.data[labels].values

        if ('ax' in mpl_kw):
            ax = mpl_kw.pop('ax')
        else:
            fig, ax = plt.subplots(1, 1, subplot_kw=dict(projection=projection))

        if discrete_labels:
            if 'markers' not in mpl_kw:
                markers = itertools.cycle(['^', 'v', '<', '>', 'o', 'd', 'p', 'x'])
            artists = []
            for label in np.unique(labels):
                mask = (labels == label)
                mpl_kw['marker'] = next(markers)
                artists.append(ax.scatter(*coords[mask].T, **mpl_kw))
        else:
            artists = ax.scatter(*coords.T, c=labels, **mpl_kw)

        if set_axes_labels:
            if projection == 'ternary':
                labels = {k: v for k, v in zip(['tlabel', 'llabel', 'rlabel'], components)}
            else:
                labels = {k: v for k, v in zip(['xlabel', 'ylabel'], components)}
            ax.set(**labels)
            ax.grid('on', color='black')
        return artists

    def plot_discrete(self, components=None, labels=None, set_labels=True, normalize=True, ternary=True, **mpl_kw):
        warnings.warn('plot_discrete is deprecated and will be removed, please use plot_scatter.', DeprecationWarning,
                      stacklevel=2)

        components = self._get_default(components)

        if ternary and len(components) == 3:
            xy = self.ternary_to_xy(components, normalize=normalize)
        elif len(components) == 2:
            xy = np.vstack(list(self.data[c].values for c in components)).T
        else:
            raise ValueError(f'Can only work with 2 or 3 components. You passed: {components}')

        if labels is None:
            if 'labels' in self.data.coords:
                labels = self.data.coords['labels'].values
            elif 'labels' in self.data:
                labels = self.data['labels'].values
            else:
                labels = np.zeros(xy.shape[0])
        elif isinstance(labels, str) and (labels in self.data):
            labels = self.data[labels].values

        artists = []
        for label in np.unique(labels):
            mask = (labels == label)
            artists.append(plt.scatter(*xy[mask].T, **mpl_kw))

        if set_labels:
            if ternary:
                mpl_format_ternary(plt.gca(), *components)
            else:
                plt.gca().set(xlabel=components[0], ylabel=components[1])
        return artists

    def plot_3D(self, components=None, labels=None, set_labels=True, **mpl_kw):
        from mpl_toolkits import mplot3d

        components = self._get_default(components)

        if not (len(components) == 3):
            raise ValueError(f'Can only work with  3 components. You passed: {components}')

        xy = np.vstack(list(self.data[c].values for c in components)).T

        if (labels is None):
            if ('labels' in self.data.coords):
                labels = self.data.coords['labels'].values
            elif ('labels' in self.data):
                labels = self.data['labels'].values
            else:
                labels = np.zeros(xy.shape[0])

        fig = plt.figure()
        ax = plt.axes(projection='3d')

        artists = []
        for label in np.unique(labels):
            mask = (labels == label)
            artists.append(ax.scatter3D(*xy[mask].T, **mpl_kw))

        ax.set(xlabel=components[0], ylabel=components[1], zlabel=components[2])

    def ternary_to_xy(self, components=None, normalize=True):
        '''Ternary composition to Cartesian coordinate'''
        components = self._get_default(components)

        if not (len(components) == 3):
            raise ValueError('Must specify exactly three components')

        comps = np.vstack(list(self.data[c].values for c in components)).T

        return ternary_to_xy(comps, normalize)

    def plot_mask(self, mask_name='mask', components_name='components_grid'):
        self.plot_discrete(self.data.attrs[components_name], labels=self.data[mask_name].astype(int), s=1)



from .Pipeline import *
from .AcquisitionFunction import *
from .Extrapolator import *
from .PairMetric import *
from .Labeler import *
from .Preprocessor import *
from .Generator import *
from .plotting import *
from .Boundary import *
from .prefab import *
from .data import *

import os
import subprocess
import warnings
from pathlib import Path

def _get_version():
    try:
        # First try to import existing version
        from ._version import __version__
        return __version__
    except ImportError:
        try:
            # Try to generate version file using hatch-vcs if we're in a git repo
            repo_root = subprocess.check_output(['git', 'rev-parse', '--show-toplevel'], 
                                             stderr=subprocess.DEVNULL).decode('utf-8').strip()
            if os.path.exists(os.path.join(repo_root, '.git')):
                try:
                    # Get version from git using similar logic to hatch-vcs
                    git_tag = subprocess.check_output(['git', 'describe', '--tags', '--always'],
                                                    stderr=subprocess.DEVNULL).decode('utf-8').strip()
                    version = git_tag.lstrip('v')
                    # Write version to file
                    version_file = Path(__file__).parent / '_version.py'
                    version_file.write_text(f'__version__ = "{version}"\n')
                    return version
                except subprocess.CalledProcessError:
                    pass
        except (subprocess.CalledProcessError, OSError):
            pass
        
        # Fallback for when we can't determine the version
        warnings.warn("Could not determine version from git or _version.py. Using default version.", ImportWarning)
        return "0.0.0+dev"

__version__ = _get_version()




import numpy as np
import matplotlib.pyplot as plt
import tqdm
from itertools import product

def ternary_to_xy(comps,normalize=True):
    '''Ternary composition to Cartesian coordinate'''
        
    if not (comps.shape[1]==3):
        raise ValueError('Must specify exactly three components')
    
    if normalize:
        comps = comps/comps.sum(1)[:,np.newaxis]
        
    # Convert ternary data to cartesian coordinates.
    xy = np.zeros((comps.shape[0],2))
    xy[:,1] = comps[:,1]*np.sin(60.*np.pi / 180.)
    xy[:,0] = comps[:,0] + xy[:,1]*np.sin(30.*np.pi/180.)/np.sin(60*np.pi/180)
    return xy
    
def mpl_format_ternary(ax=None,label_right=None,label_top=None,label_left=None):
    if ax is None:
        ax = plt.gca()
        
    ax.axis('off')
    ax.set_aspect('equal','box')
    ax.set(xlim = [0,1],ylim = [0,1])
    ax.plot([0,1,0.5,0],[0,0,np.sqrt(3)/2,0],ls='-',color='k')
    
    if label_right is not None:
        ax.text(1,0,label_right,ha='left')
    if label_top is not None:
        ax.text(0.5,np.sqrt(3)/2,label_top,ha='center',va='bottom')
    if label_left is not None:
        ax.text(0,0,label_left,ha='right')
        


def composition_grid_ternary(pts_per_row=50,basis=1.0,dim=3,eps=1e-9):
    try:
        from tqdm.contrib.itertools import product
    except ImportError:
        from itertools import product
    pts = []
    for i in product(*[np.linspace(0,1.0,pts_per_row)]*(dim-1)):
        if sum(i)>(1.0+eps):
            continue
            
        j = 1.0-sum(i)
        
        if j<(0.0-eps):
            continue
        pt = [k*basis for k in [*i,j]]
        pts.append(pt)
    return np.array(pts)

def composition_grid(components,range_spec):
    try:
        from tqdm.contrib.itertools import product
    except ImportError:
        from itertools import product
        
    grid_list =[]
    for component in components:
        spec = range_spec[component]
        grid_list.append(np.linspace(spec['min'],spec['max'],spec['steps']))
    
    pts = np.array(list(product(*grid_list)))
    return pts


"""PipelineOps for Data Preprocessing

This module contains preprocessing operations that transform, normalize, and prepare data for analysis.
Preprocessors handle tasks such as:
- Scaling and normalizing data
- Transforming between coordinate systems
- Filtering and smoothing signals
- Extracting features from raw measurements
- Converting between different data representations

Each preprocessor is implemented as a PipelineOp that can be composed with others in a processing pipeline.
"""

import warnings
from numbers import Number
from typing import Union, Optional, List, Dict

import numpy as np
import sympy
import xarray as xr
from scipy.signal import savgol_filter  # type: ignore
from typing_extensions import Self

from AFL.double_agent.PipelineOp import PipelineOp
from AFL.double_agent.util import listify


class Preprocessor(PipelineOp):
    """Base class stub for all preprocessors

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    name : str
        The name to use when added to a Pipeline. This name is used when calling Pipeline.search()
    """

    def __init__(
        self,
        input_variable: str = None,
        output_variable: str = None,
        name: str = "PreprocessorBase",
    ) -> None:
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        return NotImplementedError(".calculate must be implemented in subclasses")  # type: ignore


class SavgolFilter(Preprocessor):
    """Smooth and take derivatives of input data via a Savitsky-Golay filter

    This `PipelineOp` cleans measurement data and takes smoothed derivatives using `scipy.signal.savgol_filter`. Below
    is a summary of the steps taken.

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    dim: str
        The dimension in the `xarray.Dataset` to apply this filter over
    xlo, xhi: Optional[Number]
        The values of the input dimension (dim, above) to trim the data to
    xlo_isel, xhi_isel: Optional[int]
        The integer indices of the input dimension (dim, above) to trim the data to
    pedestal: Optional[Number]
        This value is added to the input_variable to establish a fixed data 'floor'
    npts: int
        The size of the grid to interpolate onto
    derivative: int
        The order of the derivative to return. If derivative=0, the data is smoothed with no derivative taken.
    window_length: int
        The width of the window used in the savgol smoothing. See `scipy.signal.savgol_filter` for more information.
    polyorder: int
        The order of polynomial used in the savgol smoothing. See `scipy.signal.savgol_filter` for more information.
    apply_log_scale: bool
        If True, the `input_variable` and associated `dim` coordinated are scaled with `numpy.log10`
    name: str
        The name to use when added to a Pipeline. This name is used when calling Pipeline.search()

    Notes
    -----
    This `PipelineOp` performs the following steps:

    1. Data is trimmed to `(xlo, xhi)` and then `(xlo_isel, xhi_isel)` in that order. The former trims the data to a
    numerical while the latter trims to integer indices. It is generally not advisable to supply both varieties and a
    warning will be raised if this is attempted.

    2. If `apply_log_scale = True`, both the `input_variable` and `dim` data will be scaled with `numpy.log10`. A new
    `xarray` dimension and coordinate will be created with the name `log_{dim}`.

    3. All duplicate data (multiple data values at the same `dim` coordinates) are removed by taking the average of the
    duplicates.

    4. If `pedestal` is specified, the pedestal value is added to the data and all NaNs are filled with the pedestal

    5. The data is interpolated onto a constant grid with `npts` values from the trimmed minimum to the trimmed maximum.  If `apply_log_scale=True`, the grid is geometrically rather than linearly spaced.

    6. All remaining NaN values are dropped along `dim`

    7. Finally, `scipy.signal.savgol_filter` is applied with the `window_length`, `polyorder`, and `derivative`
    parameters specified in the constructor.

    """

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        dim: str = "q",
        xlo: Optional[Number] = None,
        xhi: Optional[Number] = None,
        xlo_isel: Optional[int] = None,
        xhi_isel: Optional[int] = None,
        pedestal: Optional[Number] = None,
        npts: int = 250,
        derivative: int = 0,
        window_length: int = 31,
        polyorder: int = 2,
        apply_log_scale: bool = True,
        name: str = "SavgolFilter",
    ) -> None:
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.dim = dim
        self.npts = npts
        self.xlo = xlo
        self.xhi = xhi
        self.xlo_isel = xlo_isel
        self.xhi_isel = xhi_isel
        self.pedestal = pedestal
        self.derivative = derivative
        self.apply_log_scale = apply_log_scale
        self.polyorder = polyorder
        self.window_length = window_length

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)

        if self.dim not in data1.coords:
            raise ValueError(
                (
                    f"""dim argument of PipelineOp '{self.name}' set to '{self.dim}' but not found in coords"""
                )
            )

        if (self.xlo is not None) or (self.xhi is not None):
            if (self.xlo_isel is not None) or (self.xhi_isel is not None):
                warnings.warn(
                    (
                        "Both value and index based indexing have been specified! "
                        "Value indexing will be applied first!"
                    ),
                    stacklevel=2,
                )

        data1 = data1.transpose(self.dim,...)

        data1 = data1.sel({self.dim: slice(self.xlo, self.xhi)})
        data1 = data1.isel({self.dim: slice(self.xlo_isel, self.xhi_isel)})

        if self.apply_log_scale:
            dim = "log_" + self.dim
            data1 = data1.where(data1 > 0.0, drop=True)
            data1 = data1.pipe(np.log10)
            data1[self.dim] = np.log10(data1[self.dim])
            data1 = data1.rename({self.dim: dim})
        else:
            dim = "sg_"+self.dim
            data1[dim] = data1[self.dim]

        # need to remove duplicate values
        data1 = data1.groupby(dim, squeeze=False).mean()

        # set minimum value of scattering to pedestal value and fill nans with this value
        if self.pedestal is not None:
            data1 += self.pedestal
            data1 = data1.where(~np.isnan(data1)).fillna(self.pedestal)
        
        # interpolate to constant lin or log(dq) grid
        x_new = np.linspace(data1[dim].min().item(), data1[dim].max().item(), self.npts)
        dx = float(x_new[1] - x_new[0])
        data1 = data1.bfill(dim).ffill(dim).interpolate_na(dim)
        data1 = data1.interp({dim: x_new})

        # filter out any q that have NaN
        data1 = data1.dropna(dim, how="any")



        # take derivative
        data1_filtered = savgol_filter(
            data1.values,
            window_length=self.window_length,
            polyorder=self.polyorder,
            delta=dx,
            axis=0,
            deriv=self.derivative,
        )

        self.output[self.output_variable] = data1.copy(data=data1_filtered).transpose(...,dim)
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Savitsky-Golay filtered data"

        return self


class SubtractMin(Preprocessor):
    """Baseline input variable by subtracting minimum value"""

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        dim: str,
        name: str = "SubtractMin",
    ) -> None:
        """
        Parameters
        ----------
        input_variable : str
            The name of the `xarray.Dataset` data variable to extract from the input dataset

        output_variable : str
            The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`

        dim: str
             The dimension to use when calculating the data minimum

        name: str
            The name to use when added to a Pipeline. This name is used when calling Pipeline.search(
        """
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.dim = dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)

        self.output[self.output_variable] = data1 - data1.min(self.dim)
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Subtracted minimum value along dimension '{self.dim}'"

        return self

class Subtract(Preprocessor):
    """Baseline input variable by subtracting a value"""

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        dim: str,
        value: float | str,
        coord_value: bool = True,
        name: str = "Subtract",
    ) -> None:
        """
        Parameters
        ----------
        input_variable : str
            The name of the `xarray.Dataset` data variable to extract from the input dataset

        output_variable : str
            The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`

        dim: str
             The dimension to use when calculating the data minimum

        name: str
            The name to use when added to a Pipeline. This name is used when calling Pipeline.search(
        """
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.dim = dim
        self.value = value
        self.coord_value = coord_value

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)
        
        if isinstance(self.value,str) and self.coord_value:
            value = dataset[self.value].item()
            subtract_value = data1.sel({self.dim:value},method='nearest')
        elif self.coord_value:
            subtract_value = data1.sel({self.dim:self.value},method='nearest')
        else:
            subtract_value = self.value
            
        self.output[self.output_variable] = data1 - subtract_value
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Subtracted minimum value along dimension '{self.dim}'"

        return self


class Standardize(Preprocessor):
    """Standardize the data to have min 0 and max 1

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    dim : str
        The dimension used for calculating the data minimum
    component_dim : str | None, default="component"
        The dimension for component-wise operations
    scale_variable : str | None, default=None
        If specified, the min/max of this data variable in the supplied `xarray.Dataset` will be used to scale the
        data rather than min/max of the `input_variable` or the supplied `min_val` or `max_val`
    min_val : Number | None, default=None
        Value used to scale the data minimum
    max_val : Number | None, default=None
        Value used to scale the data maximum
    name : str, default="Standardize"
        The name to use when added to a Pipeline
    """

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        dim: str,
        component_dim: str | None = "component",
        scale_variable: str | None = None,
        min_val: Number | None = None,
        max_val: Number | None = None,
        name: str = "Standardize",
    ) -> None:
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.dim = dim
        self.component_dim = component_dim
        self.min_val = min_val
        self.max_val = max_val
        self.scale_variable = scale_variable

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)

        if self.max_val is None:
            if self.scale_variable is None:
                max_val = data1.max(self.dim)
            else:
                max_val = dataset[self.scale_variable].max(self.dim)
        elif isinstance(self.max_val,dict):
            max_val = xr.Dataset(self.max_val).to_array(self.component_dim)
        else:
            max_val = self.max_val

        if self.min_val is None:
            if self.scale_variable is None:
                min_val = data1.min(self.dim)
            else:
                min_val = dataset[self.scale_variable].min(self.dim)
        elif isinstance(self.min_val, dict):
            min_val = xr.Dataset(self.min_val).to_array(self.component_dim)
        else:
            min_val = self.min_val

        self.output[self.output_variable] = (data1 - min_val) / (max_val - min_val)
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Data normalized to have range 0 -> 1"

        return self


class Destandardize(Preprocessor):
    """Transform the data from 0->1 scaling

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    dim : str
        The dimension used for calculating the data minimum
    component_dim : str | None, default="component"
        The dimension for component-wise operations
    scale_variable : str | None, default=None
        If specified, the min/max of this data variable in the supplied `xarray.Dataset` will be used to scale the
        data rather than min/max of the `input_variable` or the supplied `min_val` or `max_val`
    min_val : Number | None, default=None
        Value used to scale the data minimum
    max_val : Number | None, default=None
        Value used to scale the data maximum
    name : str, default="Destandardize"
        The name to use when added to a Pipeline
    """

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        dim: str,
        component_dim: str | None = "component",
        scale_variable: str | None = None,
        min_val: Number | None = None,
        max_val: Number | None = None,
        name: str = "Destandardize",
    ) -> None:

        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.dim = dim
        self.component_dim = component_dim
        self.min_val = min_val
        self.max_val = max_val
        self.scale_variable = scale_variable

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)

        if self.max_val is None:
            if self.scale_variable is None:
                max_val = data1.max(self.dim)
            else:
                max_val = dataset[self.scale_variable].max(self.dim)
        elif isinstance(self.max_val,dict):
            max_val = xr.Dataset(self.max_val).to_array(self.component_dim)
        else:
            max_val = self.max_val

        if self.min_val is None:
            if self.scale_variable is None:
                min_val = data1.min(self.dim)
            else:
                min_val = dataset[self.scale_variable].min(self.dim)
        elif isinstance(self.min_val, dict):
            min_val = xr.Dataset(self.min_val).to_array(self.component_dim)
        else:
            min_val = self.min_val

        self.output[self.output_variable] = data1 * (max_val - min_val) + min_val
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Data normalized to have range 0 -> 1"

        return self


class Zscale(Preprocessor):
    """Z-scale the data to have mean 0 and standard deviation scaling

        Parameters
        ----------
        input_variable : str
            The name of the `xarray.Dataset` data variable to extract from the input dataset

        output_variable : str
            The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`

        dim: str
             The dimension to use when calculating the data minimum

        name: str
            The name to use when added to a Pipeline. This name is used when calling Pipeline.search(
        """

    def __init__(
        self, input_variable: str, output_variable: str, dim: str, name: str = "Zscale"
    ) -> None:
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.mean = None
        self.std = None
        self.dim = dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)
        self.mean = data1.mean()
        self.std = data1.std()

        # this standardizes between the min and the max values, this isn't z-scaling...
        self.output[self.output_variable] = (data1 - self.mean) / self.std
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Data Zscaled to have range 0 ~ -1 -> 1"

        return self


class ZscaleError(Preprocessor):
    """Scale the y_err data, first input is y, second input is y_err

    Parameters
    ----------
    input_variables : Union[Optional[str], List[str]]
        The names of the input variables - first is y, second is y_err
    output_variable : str
        The name of the variable to be inserted into the dataset
    dim : str
        The dimension to use when calculating the data minimum
    name : str, default="Zscale_error"
        The name to use when added to a Pipeline
    """

    def __init__(
        self,
        input_variables: Union[Optional[str], List[str]],
        output_variable: str,
        dim: str,
        name: str = "Zscale_error",
    ) -> None:
        super().__init__(
            name=name, input_variable=input_variables, output_variable=output_variable
        )

        self.dim = dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data1 = self._get_variable(dataset)
        data_y_err = data1[self.input_variable[1]]
        data_y = data1[self.input_variable[0]]

        self.output[self.output_variable] = data_y_err / data_y.std()
        self.output[self.output_variable].attrs[
            "description"
        ] = f"Uncertainty normalized to have range 0 -> inf"

        return self


class BarycentricToTernaryXY(Preprocessor):
    """
    Transform from ternary coordinates to xy coordinates

    Note ---- Adapted from BaryCentric transform mpltern:
    https://github.com/yuzie007/mpltern/blob/master/mpltern/ternary/transforms.py

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset

    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`

    sample_dim: str
         The dimension to use when calculating the data minimum

    name: str
        The name to use when added to a Pipeline. This name is used when calling Pipeline.search(
    """

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        name: str = "BarycentricToTernaryXY",
    ) -> None:
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        x_min = 0.5 - 1.0 / np.sqrt(3)
        x_max = 0.5 + 1.0 / np.sqrt(3)
        self.corners = np.array([(0.5, 1.0), (x_min, 0.0), (x_max, 0.0)])

        self.sample_dim = sample_dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        bary = self._get_variable(dataset)
        bary = bary.transpose(self.sample_dim, ...).values
        bary = bary / np.sum(bary, axis=1)[:, np.newaxis]
        xy = np.dot(bary, self.corners)

        self.output[self.output_variable] = xr.DataArray(
            xy, dims=[self.sample_dim, "xy"]
        )
        self.output[self.output_variable].attrs["description"] = (
            "barycentric coordinates from xy variable " f"'{self.input_variable}'"
        )
        return self


class TernaryXYToBarycentric(Preprocessor):
    """Transform to ternary coordinates from xy coordinates

    Note 
    ---- 
    Adapted from BaryCentric transform mpltern:
    https://github.com/yuzie007/mpltern/blob/master/mpltern/ternary/transforms.py

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    sample_dim : str
        The dimension to use when calculating the data minimum
    name : str, default="TernaryXYToBarycentric"
        The name to use when added to a Pipeline
    """

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        name: str = "TernaryXYToBarycentric",
    ) -> None:
        """
        Parameters
        ----------
        input_variable : str
            The name of the `xarray.Dataset` data variable to extract from the input dataset

        output_variable : str
            The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`

        sample_dim: str
             The dimension to use when calculating the data minimum

        name: str
            The name to use when added to a Pipeline. This name is used when calling Pipeline.search(
        """
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        x_min = 0.5 - 1.0 / np.sqrt(3)
        x_max = 0.5 + 1.0 / np.sqrt(3)
        self.corners = [(0.5, 1.0), (x_min, 0.0), (x_max, 0.0)]

        self.sample_dim = sample_dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        xy = self._get_variable(dataset)
        xy = xy.transpose(self.sample_dim, ...).values
        xys = np.column_stack((xy, np.ones(xy.shape[0])))
        v = np.column_stack((self.corners, np.ones(3)))
        bary = np.dot(xys, np.linalg.inv(v))
        self.output[self.output_variable] = xr.DataArray(
            bary, dims=[self.sample_dim, "component"]
        )
        self.output[self.output_variable].attrs["description"] = (
            "XY coordinates from barycentric variable " f"'{self.input_variable}'"
        )
        return self


class SympyTransform(Preprocessor):
    """Transform data using sympy expressions

    Parameters
    ----------
    input_variable : str
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : str
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    sample_dim : str
        The sample dimension i.e., the dimension of compositions or grid points
    component_dim : str, default="component"
        The dimension of the component of each gridpoint
    transforms : Dict[str,object]
        A dictionary of transforms (sympy expressions) to evaluate to generate new variables. For this method to
        function, the transforms must be completely specified except for the names in component_dim of the
        input_variable
    transform_dim : str
        The name of the dimension that the 'component_dim' will be transformed to
    name : str, default="SympyTransform"
        The name to use when added to a Pipeline

    Example
    -------
    ```python
    from AFL.double_agent import *
    import sympy
    with Pipeline() as p:
        CartesianGrid(
            output_variable='comps',
            grid_spec={
                'A':{'min':1,'max':25,'steps':5},
                'B':{'min':1,'max':25,'steps':5},
                'C':{'min':1,'max':25,'steps':5},
            },
            sample_dim='grid'
        )

        A,B,C = sympy.symbols('A B C')
        vA = A/(A+B+C)
        vB = B/(A+B+C)
        vC = C/(A+B+C)
        SympyTransform(
            input_variable='comps',
            output_variable='trans_comps',
            sample_dim='grid',
            transforms={'vA':vA,'vB':vB,'vC':vC},
            transform_dim='trans_component'
        )

    p.calculate(xr.Dataset())# returns dataset with grid and transformed grid
    ```
    """

    def __init__(
        self,
        input_variable: str,
        output_variable: str,
        sample_dim: str,
        transforms: Dict[str, object],
        transform_dim: str,
        component_dim: str = "component",
        name: str = "SympyTransform",
    ) -> None:

        # must convert to strings for JSON serialization
        transforms = {k:str(v) for k,v in transforms.items()}

        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )
        self.sample_dim = sample_dim
        self.component_dim = component_dim
        self.transforms = transforms
        self.transform_dim = transform_dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data = dataset[self.input_variable].transpose(
            self.sample_dim, self.component_dim
        )

        # need to construct a dict of arrays
        comps = {
            k: v.squeeze().values
            for k, v in data.groupby(self.component_dim, squeeze=False)
        }

        # apply transform
        new_comps = xr.Dataset()
        for name, transform in self.transforms.items():
            transform = sympy.sympify(transform)
            symbols = list(transform.free_symbols)
            lam = sympy.lambdify(symbols, transform)
            new_comps[name] = (
                (self.sample_dim,),
                listify(lam(**{k.name: comps[k.name] for k in symbols})),
            )

        new_comps = new_comps.to_array(self.transform_dim).transpose(
            ..., self.transform_dim
        )

        self.output[self.output_variable] = new_comps
        self.output[self.output_variable].attrs[
            "description"
        ] = "Variables transformed using sympy expressions"

        return self


class Extrema(Preprocessor):
    """Find the extrema of a data variable"""

    def __init__(
            self,
            input_variable: str,
            output_variable: str,
            dim: str,
            return_coords: bool = False,
            operator='max',
            slice: Optional[List] = None,
            slice_dim: Optional[str] = None,
            name: str = "Extrema",
    ) -> None:
        """
        Parameters
        ----------
        input_variable : str
            The name of the `xarray.Dataset` data variable to extract from the input dataset

        output_variable : str
            The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`

        dim: str
             The dimension to use when calculating the data minimum

        name: str
            The name to use when added to a Pipeline. This name is used when calling Pipeline.search(
        """
        super().__init__(
            name=name, input_variable=input_variable, output_variable=output_variable
        )

        self.dim = dim
        self.operator = operator
        self.return_coords = return_coords
        self.slice = slice
        self.slice_dim = slice_dim

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Apply this `PipelineOp` to the supplied `xarray.dataset`"""
        data = self._get_variable(dataset)

        if self.slice is not None:
            data = data.sel({self.slice_dim: slice(*self.slice)})

        if self.return_coords:
            if 'arg' not in self.operator:
                operator = 'arg' + self.operator
            else:
                operator = self.operator
            idx = getattr(data, operator)(dim=self.dim)
            self.output[self.output_variable] = data[self.dim][idx]
        else:
            self.output[self.output_variable] = getattr(data, self.operator)(dim=self.dim)

        self.output[self.output_variable].attrs[
            "description"
        ] = f"Extrema of {self.input_variable} with operator {self.operator} and return_coords {self.return_coords}"

        return self

class VarsToArray(Preprocessor):
    """Convert multiple variables into a single array

    Parameters
    ----------
    input_variables : List
        List of input variables to combine into an array
    output_variable : str
        The name of the variable to be inserted into the dataset
    variable_dim : str
        The dimension name for the variables in the output array
    squeeze : bool, default=False
        Whether to squeeze out single-dimension axes
    variable_mapping : Dict, default=None
        Optional mapping to rename variables
    name : str, default='VarsToArray'
        The name to use when added to a Pipeline
    """

    def __init__(
        self, 
        input_variables: List, 
        output_variable: str, 
        variable_dim: str,
        squeeze: bool = False,
        variable_mapping: Dict = None,
        name: str = 'VarsToArray'):

        super().__init__(name=name, input_variable=input_variables, output_variable=output_variable)
        print(self.input_variable,self.output_variable)

        if variable_mapping is None:
            self.variable_mapping = {}
        else:
            self.variable_mapping = variable_mapping
        
        self.variable_dim = variable_dim
        self.squeeze = squeeze

    def calculate(self, dataset):
        output = dataset[self.input_variable].rename_vars(self.variable_mapping).to_array(self.variable_dim)      
        if self.squeeze:
            output = output.squeeze()
        print(self.output_variable)
        print(output)
        self.output[self.output_variable] = output
        return self

class ArrayToVars(Preprocessor):
    """Convert an array into multiple variables

    Parameters
    ----------
    input_variable : str
        The name of the array variable to split into separate variables
    output_variables : list
        The names of the variables to create from the array
    split_dim : str
        The dimension to split the array along
    postfix : str, default=''
        String to append to output variable names
    squeeze : bool, default=False
        Whether to squeeze out single-dimension axes
    name : str, default='DatasetToVars'
        The name to use when added to a Pipeline
    """

    def __init__(
        self, 
        input_variable: str, 
        output_variables: list, 
        split_dim: list,
        postfix: str = '',
        squeeze: bool = False,
        name: str = 'DatasetToVars'):

        super().__init__(name=name, input_variable=input_variable, output_variable=output_variables)
        print(self.input_variable,self.output_variable)
        
        self.split_dim = split_dim
        self.squeeze = squeeze
        self.output_variables = output_variables
        self.postfix= postfix
    def calculate(self, dataset):
        input_arr = dataset[self.input_variable]   
        if self.squeeze:
            output = output.squeeze()
        for var in self.output_variables:
            print(var)
    
            self.output[var+self.postfix] = input_arr.sel({self.split_dim:var})
            
        return self

